# -*- encoding: utf-8 -*-
'''Display functions'''

from collections import defaultdict

import numpy as np
from scipy.signal import spectrogram

from matplotlib.patches import Rectangle
from matplotlib.ticker import FuncFormatter, MultipleLocator
from matplotlib.ticker import Formatter
from matplotlib.colors import LinearSegmentedColormap, LogNorm, ColorConverter
from matplotlib.collections import BrokenBarHCollection

from .melody import freq_to_voicing
from .util import midi_to_hz, hz_to_midi


def __expand_limits(ax, limits, which='x'):
    '''Helper function to expand axis limits'''

    if which == 'x':
        getter, setter = ax.get_xlim, ax.set_xlim
    elif which == 'y':
        getter, setter = ax.get_ylim, ax.set_ylim
    else:
        raise ValueError('invalid axis: {}'.format(which))

    old_lims = getter()
    new_lims = list(limits)

    # infinite limits occur on new axis objects with no data
    if np.isfinite(old_lims[0]):
        new_lims[0] = min(old_lims[0], limits[0])

    if np.isfinite(old_lims[1]):
        new_lims[1] = max(old_lims[1], limits[1])

    setter(new_lims)


def __get_axes(ax=None, fig=None):
    '''Get or construct the target axes object for a new plot.

    Parameters
    ----------
    ax : matplotlib.pyplot.axes, optional
        If provided, return this axes object directly.

    fig : matplotlib.figure.Figure, optional
        The figure to query for axes.

        By default, uses the current figure `plt.gcf()`.

    Returns
    -------
    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the segmentation.
        If none is provided, a new set of axes is created.

    new_axes : bool
        If `True`, the axis object was newly constructed.
        If `False`, the axis object already existed.

    '''

    new_axes = False

    if ax is not None:
        return ax, new_axes

    if fig is None:
        import matplotlib.pyplot as plt
        fig = plt.gcf()

    if not fig.get_axes():
        new_axes = True

    return fig.gca(), new_axes


def segments(intervals, labels, base=None, height=None, text=False,
             text_kw=None, ax=None, **kwargs):
    '''Plot a segmentation as a set of disjoint rectangles.

    Parameters
    ----------
    intervals : np.ndarray, shape=(n, 2)
        segment intervals, in the format returned by
        :func:`mir_eval.io.load_intervals` or
        :func:`mir_eval.io.load_labeled_intervals`.

    labels : list, shape=(n,)
        reference segment labels, in the format returned by
        :func:`mir_eval.io.load_labeled_intervals`.

    base : number
        The vertical position of the base of the rectangles.
        By default, this will be the bottom of the plot.

    height : number
        The height of the rectangles.
        By default, this will be the top of the plot (minus ``base``).

    text : bool
        If true, each segment's label is displayed in its
        upper-left corner

    text_kw : dict
        If ``text == True``, the properties of the text
        object can be specified here.
        See ``matplotlib.pyplot.Text`` for valid parameters

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the segmentation.
        If none is provided, a new set of axes is created.

    kwargs
        Additional keyword arguments to pass to
        ``matplotlib.patches.Rectangle``.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''
    if text_kw is None:
        text_kw = dict()
    text_kw.setdefault('va', 'top')
    text_kw.setdefault('clip_on', True)
    text_kw.setdefault('bbox', dict(boxstyle='round', facecolor='white'))

    # Make sure we have a numpy array
    intervals = np.atleast_2d(intervals)

    seg_def_style = dict(linewidth=1)

    ax, new_axes = __get_axes(ax=ax)

    if new_axes:
        ax.set_ylim([0, 1])

    # Infer height
    if base is None:
        base = ax.get_ylim()[0]

    if height is None:
        height = ax.get_ylim()[1]

    cycler = ax._get_patches_for_fill.prop_cycler

    seg_map = dict()

    for lab in labels:
        if lab in seg_map:
            continue

        style = next(cycler)
        seg_map[lab] = seg_def_style.copy()
        seg_map[lab].update(style)
        # Swap color -> facecolor here so we preserve edgecolor on rects
        seg_map[lab]['facecolor'] = seg_map[lab].pop('color')
        seg_map[lab].update(kwargs)
        seg_map[lab]['label'] = lab

    for ival, lab in zip(intervals, labels):
        rect = Rectangle((ival[0], base), ival[1] - ival[0], height,
                         **seg_map[lab])
        ax.add_patch(rect)
        seg_map[lab].pop('label', None)

        if text:
            ann = ax.annotate(lab,
                              xy=(ival[0], height), xycoords='data',
                              xytext=(8, -10), textcoords='offset points',
                              **text_kw)
            ann.set_clip_path(rect)

    if new_axes:
        ax.set_yticks([])

    # Only expand if we have data
    if intervals.size:
        __expand_limits(ax, [intervals.min(), intervals.max()], which='x')

    return ax


def labeled_intervals(intervals, labels, label_set=None,
                      base=None, height=None, extend_labels=True,
                      ax=None, tick=True, **kwargs):
    '''Plot labeled intervals with each label on its own row.

    Parameters
    ----------
    intervals : np.ndarray, shape=(n, 2)
        segment intervals, in the format returned by
        :func:`mir_eval.io.load_intervals` or
        :func:`mir_eval.io.load_labeled_intervals`.

    labels : list, shape=(n,)
        reference segment labels, in the format returned by
        :func:`mir_eval.io.load_labeled_intervals`.

    label_set : list
        An (ordered) list of labels to determine the plotting order.
        If not provided, the labels will be inferred from
        ``ax.get_yticklabels()``.
        If no ``yticklabels`` exist, then the sorted set of unique values
        in ``labels`` is taken as the label set.

    base : np.ndarray, shape=(n,), optional
        Vertical positions of each label.
        By default, labels are positioned at integers
        ``np.arange(len(labels))``.

    height : scalar or np.ndarray, shape=(n,), optional
        Height for each label.
        If scalar, the same value is applied to all labels.
        By default, each label has ``height=1``.

    extend_labels : bool
        If ``False``, only values of ``labels`` that also exist in
        ``label_set`` will be shown.

        If ``True``, all labels are shown, with those in `labels` but
        not in `label_set` appended to the top of the plot.
        A horizontal line is drawn to indicate the separation between
        values in or out of ``label_set``.

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the intervals.
        If none is provided, a new set of axes is created.

    tick : bool
        If ``True``, sets tick positions and labels on the y-axis.

    kwargs
        Additional keyword arguments to pass to
        `matplotlib.collection.BrokenBarHCollection`.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''

    # Get the axes handle
    ax, _ = __get_axes(ax=ax)

    # Make sure we have a numpy array
    intervals = np.atleast_2d(intervals)

    if label_set is None:
        # If we have non-empty pre-existing tick labels, use them
        label_set = [_.get_text() for _ in ax.get_yticklabels()]
        # If none of the label strings have content, treat it as empty
        if not any(label_set):
            label_set = []
    else:
        label_set = list(label_set)

    # Put additional labels at the end, in order
    if extend_labels:
        ticks = label_set + sorted(set(labels) - set(label_set))
    elif label_set:
        ticks = label_set
    else:
        ticks = sorted(set(labels))

    style = dict(linewidth=1)

    style.update(next(ax._get_patches_for_fill.prop_cycler))
    # Swap color -> facecolor here so we preserve edgecolor on rects
    style['facecolor'] = style.pop('color')
    style.update(kwargs)

    if base is None:
        base = np.arange(len(ticks))

    if height is None:
        height = 1

    if np.isscalar(height):
        height = height * np.ones_like(base)

    seg_y = dict()
    for ybase, yheight, lab in zip(base, height, ticks):
        seg_y[lab] = (ybase, yheight)

    xvals = defaultdict(list)
    for ival, lab in zip(intervals, labels):
        if lab not in seg_y:
            continue
        xvals[lab].append((ival[0], ival[1] - ival[0]))

    for lab in seg_y:
        ax.add_collection(BrokenBarHCollection(xvals[lab], seg_y[lab],
                                               **style))
        # Pop the label after the first time we see it, so we only get
        # one legend entry
        style.pop('label', None)

    # Draw a line separating the new labels from pre-existing labels
    if label_set != ticks:
        ax.axhline(len(label_set), color='k', alpha=0.5)

    if tick:
        ax.grid(True, axis='y')
        ax.set_yticks([])
        ax.set_yticks(base)
        ax.set_yticklabels(ticks, va='bottom')
        ax.yaxis.set_major_formatter(IntervalFormatter(base, ticks))

    if base.size:
        __expand_limits(ax, [base.min(), (base + height).max()], which='y')
    if intervals.size:
        __expand_limits(ax, [intervals.min(), intervals.max()], which='x')

    return ax


class IntervalFormatter(Formatter):
    '''Ticker formatter for labeled interval plots.

    Parameters
    ----------
    base : array-like of int
        The base positions of each label

    ticks : array-like of string
        The labels for the ticks
    '''
    def __init__(self, base, ticks):

        self._map = {int(k): v for k, v in zip(base, ticks)}

    def __call__(self, x, pos=None):

        return self._map.get(int(x), '')


def hierarchy(intervals_hier, labels_hier, levels=None, ax=None, **kwargs):
    '''Plot a hierarchical segmentation

    Parameters
    ----------
    intervals_hier : list of np.ndarray
        A list of segmentation intervals.  Each element should be
        an n-by-2 array of segment intervals, in the format returned by
        :func:`mir_eval.io.load_intervals` or
        :func:`mir_eval.io.load_labeled_intervals`.
        Segmentations should be ordered by increasing specificity.

    labels_hier : list of list-like
        A list of segmentation labels.  Each element should
        be a list of labels for the corresponding element in
        `intervals_hier`.

    levels : list of string
        Each element ``levels[i]`` is a label for the ```i`` th segmentation.
        This is used in the legend to denote the levels in a segment hierarchy.

    kwargs
        Additional keyword arguments to `labeled_intervals`.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''

    # This will break if a segment label exists in multiple levels
    if levels is None:
        levels = list(range(len(intervals_hier)))

    # Get the axes handle
    ax, _ = __get_axes(ax=ax)

    # Count the pre-existing patches
    n_patches = len(ax.patches)

    for ints, labs, key in zip(intervals_hier[::-1],
                               labels_hier[::-1],
                               levels[::-1]):
        labeled_intervals(ints, labs, label=key, ax=ax, **kwargs)

    # Reverse the patch ordering for anything we've added.
    # This way, intervals are listed in the legend from top to bottom
    ax.patches[n_patches:] = ax.patches[n_patches:][::-1]
    return ax


def events(times, labels=None, base=None, height=None, ax=None, text_kw=None,
           **kwargs):
    '''Plot event times as a set of vertical lines

    Parameters
    ----------
    times : np.ndarray, shape=(n,)
        event times, in the format returned by
        :func:`mir_eval.io.load_events` or
        :func:`mir_eval.io.load_labeled_events`.

    labels : list, shape=(n,), optional
        event labels, in the format returned by
        :func:`mir_eval.io.load_labeled_events`.

    base : number
        The vertical position of the base of the line.
        By default, this will be the bottom of the plot.

    height : number
        The height of the lines.
        By default, this will be the top of the plot (minus `base`).

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the segmentation.
        If none is provided, a new set of axes is created.

    text_kw : dict
        If `labels` is provided, the properties of the text
        objects can be specified here.
        See `matplotlib.pyplot.Text` for valid parameters

    kwargs
        Additional keyword arguments to pass to
        `matplotlib.pyplot.vlines`.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''
    if text_kw is None:
        text_kw = dict()
    text_kw.setdefault('va', 'top')
    text_kw.setdefault('clip_on', True)
    text_kw.setdefault('bbox', dict(boxstyle='round', facecolor='white'))

    # make sure we have an array for times
    times = np.asarray(times)

    # Get the axes handle
    ax, new_axes = __get_axes(ax=ax)

    # If we have fresh axes, set the limits

    if new_axes:
        # Infer base and height
        if base is None:
            base = 0
        if height is None:
            height = 1

        ax.set_ylim([base, height])
    else:
        if base is None:
            base = ax.get_ylim()[0]

        if height is None:
            height = ax.get_ylim()[1]

    cycler = ax._get_patches_for_fill.prop_cycler

    style = next(cycler).copy()
    style.update(kwargs)
    # If the user provided 'colors', don't override it with 'color'
    if 'colors' in style:
        style.pop('color', None)

    lines = ax.vlines(times, base, base + height, **style)

    if labels:
        for path, lab in zip(lines.get_paths(), labels):
            ax.annotate(lab,
                        xy=(path.vertices[0][0], height),
                        xycoords='data',
                        xytext=(8, -10), textcoords='offset points',
                        **text_kw)

    if new_axes:
        ax.set_yticks([])

    __expand_limits(ax, [base, base + height], which='y')

    if times.size:
        __expand_limits(ax, [times.min(), times.max()], which='x')

    return ax


def pitch(times, frequencies, midi=False, unvoiced=False, ax=None, **kwargs):
    '''Visualize pitch contours

    Parameters
    ----------
    times : np.ndarray, shape=(n,)
        Sample times of frequencies

    frequencies : np.ndarray, shape=(n,)
        frequencies (in Hz) of the pitch contours.
        Voicing is indicated by sign (positive for voiced,
        non-positive for non-voiced).

    midi : bool
        If `True`, plot on a MIDI-numbered vertical axis.
        Otherwise, plot on a linear frequency axis.

    unvoiced : bool
        If `True`, unvoiced pitch contours are plotted and indicated
        by transparency.

        Otherwise, unvoiced pitch contours are omitted from the display.

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the pitch contours.
        If none is provided, a new set of axes is created.

    kwargs
        Additional keyword arguments to `matplotlib.pyplot.plot`.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''

    ax, _ = __get_axes(ax=ax)

    times = np.asarray(times)

    # First, segment into contiguously voiced contours
    frequencies, voicings = freq_to_voicing(np.asarray(frequencies,
                                                       dtype=np.float))
    voicings = voicings.astype(bool)

    # Here are all the change-points
    v_changes = 1 + np.flatnonzero(voicings[1:] != voicings[:-1])
    v_changes = np.unique(np.concatenate([[0], v_changes, [len(voicings)]]))

    # Set up arrays of slices for voiced and unvoiced regions
    v_slices, u_slices = [], []
    for start, end in zip(v_changes, v_changes[1:]):
        idx = slice(start, end)
        # A region is voiced if its starting sample is voiced
        # It's unvoiced if none of the samples in the region are voiced.
        if voicings[start]:
            v_slices.append(idx)
        elif frequencies[idx].all():
            u_slices.append(idx)

    # Now we just need to plot the contour
    style = dict()
    style.update(next(ax._get_lines.prop_cycler))
    style.update(kwargs)

    if midi:
        idx = frequencies > 0
        frequencies[idx] = hz_to_midi(frequencies[idx])

        # Tick at integer midi notes
        ax.yaxis.set_minor_locator(MultipleLocator(1))

    for idx in v_slices:
        ax.plot(times[idx], frequencies[idx], **style)
        style.pop('label', None)

    # Plot the unvoiced portions
    if unvoiced:
        style['alpha'] = style.get('alpha', 1.0) * 0.5
        for idx in u_slices:
            ax.plot(times[idx], frequencies[idx], **style)

    return ax


def multipitch(times, frequencies, midi=False, unvoiced=False, ax=None,
               **kwargs):
    '''Visualize multiple f0 measurements

    Parameters
    ----------
    times : np.ndarray, shape=(n,)
        Sample times of frequencies

    frequencies : list of np.ndarray
        frequencies (in Hz) of the pitch measurements.
        Voicing is indicated by sign (positive for voiced,
        non-positive for non-voiced).

        `times` and `frequencies` should be in the format produced by
        :func:`mir_eval.io.load_ragged_time_series`

    midi : bool
        If `True`, plot on a MIDI-numbered vertical axis.
        Otherwise, plot on a linear frequency axis.

    unvoiced : bool
        If `True`, unvoiced pitches are plotted and indicated
        by transparency.

        Otherwise, unvoiced pitches are omitted from the display.

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the pitch contours.
        If none is provided, a new set of axes is created.

    kwargs
        Additional keyword arguments to `plt.scatter`.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''

    # Get the axes handle
    ax, _ = __get_axes(ax=ax)

    # Set up a style for the plot
    style_voiced = dict()
    style_voiced.update(next(ax._get_lines.prop_cycler))
    style_voiced.update(kwargs)

    style_unvoiced = style_voiced.copy()
    style_unvoiced.pop('label', None)
    style_unvoiced['alpha'] = style_unvoiced.get('alpha', 1.0) * 0.5

    # We'll collect all times and frequencies first, then plot them
    voiced_times = []
    voiced_freqs = []

    unvoiced_times = []
    unvoiced_freqs = []

    for t, freqs in zip(times, frequencies):
        if not len(freqs):
            continue

        freqs, voicings = freq_to_voicing(np.asarray(freqs, dtype=np.float))

        # Discard all 0-frequency measurements
        idx = freqs > 0
        freqs = freqs[idx]
        voicings = voicings[idx].astype(bool)

        if midi:
            freqs = hz_to_midi(freqs)

        n_voiced = sum(voicings)
        voiced_times.extend([t] * int(n_voiced))
        voiced_freqs.extend(freqs[voicings])
        unvoiced_times.extend([t] * (len(freqs) - n_voiced))
        unvoiced_freqs.extend(freqs[~voicings])

    # Plot the voiced frequencies
    ax.scatter(voiced_times, voiced_freqs, **style_voiced)

    # Plot the unvoiced frequencies
    if unvoiced:
        ax.scatter(unvoiced_times, unvoiced_freqs, **style_unvoiced)

    # Tick at integer midi notes
    if midi:
        ax.yaxis.set_minor_locator(MultipleLocator(1))
    return ax


def piano_roll(intervals, pitches=None, midi=None, ax=None, **kwargs):
    '''Plot a quantized piano roll as intervals

    Parameters
    ----------
    intervals : np.ndarray, shape=(n, 2)
        timing intervals for notes

    pitches : np.ndarray, shape=(n,), optional
        pitches of notes (in Hz).

    midi : np.ndarray, shape=(n,), optional
        pitches of notes (in MIDI numbers).

        At least one of ``pitches`` or ``midi`` must be provided.

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the intervals.
        If none is provided, a new set of axes is created.

    kwargs
        Additional keyword arguments to :func:`labeled_intervals`.

    Returns
    -------
    ax : matplotlib.pyplot.axes._subplots.AxesSubplot
        A handle to the (possibly constructed) plot axes
    '''

    if midi is None:
        if pitches is None:
            raise ValueError('At least one of `midi` or `pitches` '
                             'must be provided.')

        midi = hz_to_midi(pitches)

    scale = np.arange(128)
    ax = labeled_intervals(intervals, np.round(midi).astype(int),
                           label_set=scale,
                           tick=False,
                           ax=ax,
                           **kwargs)

    # Minor tick at each semitone
    ax.yaxis.set_minor_locator(MultipleLocator(1))

    ax.axis('auto')
    return ax


def separation(sources, fs=22050, labels=None, alpha=0.75, ax=None, **kwargs):
    '''Source-separation visualization

    Parameters
    ----------
    sources : np.ndarray, shape=(nsrc, nsampl)
        A list of waveform buffers corresponding to each source

    fs : number > 0
        The sampling rate

    labels : list of strings
        An optional list of descriptors corresponding to each source

    alpha : float in [0, 1]
        Maximum alpha (opacity) of spectrogram values.

    ax : matplotlib.pyplot.axes
        An axis handle on which to draw the spectrograms.
        If none is provided, a new set of axes is created.

    kwargs
        Additional keyword arguments to ``scipy.signal.spectrogram``

    Returns
    -------
    ax
        The axis handle for this plot
    '''

    # Get the axes handle
    ax, new_axes = __get_axes(ax=ax)

    # Make sure we have at least two dimensions
    sources = np.atleast_2d(sources)

    if labels is None:
        labels = ['Source {:d}'.format(_) for _ in range(len(sources))]

    kwargs.setdefault('scaling', 'spectrum')

    # The cumulative spectrogram across sources
    # is used to establish the reference power
    # for each individual source
    cumspec = None
    specs = []
    for i, src in enumerate(sources):
        freqs, times, spec = spectrogram(src, fs=fs, **kwargs)

        specs.append(spec)
        if cumspec is None:
            cumspec = spec.copy()
        else:
            cumspec += spec

    ref_max = cumspec.max()
    ref_min = ref_max * 1e-6

    color_conv = ColorConverter()

    for i, spec in enumerate(specs):

        # For each source, grab a new color from the cycler
        # Then construct a colormap that interpolates from
        # [transparent white -> new color]
        color = next(ax._get_lines.prop_cycler)['color']
        color = color_conv.to_rgba(color, alpha=alpha)
        cmap = LinearSegmentedColormap.from_list(labels[i],
                                                 [(1.0, 1.0, 1.0, 0.0),
                                                  color])

        ax.pcolormesh(times, freqs, spec,
                      cmap=cmap,
                      norm=LogNorm(vmin=ref_min, vmax=ref_max),
                      shading='gouraud',
                      label=labels[i])

        # Attach a 0x0 rect to the axis with the corresponding label
        # This way, it will show up in the legend
        ax.add_patch(Rectangle((0, 0), 0, 0, color=color, label=labels[i]))

    if new_axes:
        ax.axis('tight')

    return ax


def __ticker_midi_note(x, pos):
    '''A ticker function for midi notes.

    Inputs x are interpreted as midi numbers, and converted
    to [NOTE][OCTAVE]+[cents].
    '''

    NOTES = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']

    cents = float(np.mod(x, 1.0))
    if cents >= 0.5:
        cents = cents - 1.0
        x = x + 0.5

    idx = int(x % 12)

    octave = int(x / 12) - 1

    if cents == 0:
        return '{:s}{:2d}'.format(NOTES[idx], octave)
    return '{:s}{:2d}{:+02d}'.format(NOTES[idx], octave, int(cents * 100))


def __ticker_midi_hz(x, pos):
    '''A ticker function for midi pitches.

    Inputs x are interpreted as midi numbers, and converted
    to Hz.
    '''

    return '{:g}'.format(midi_to_hz(x))


def ticker_notes(ax=None):
    '''Set the y-axis of the given axes to MIDI notes

    Parameters
    ----------
    ax : matplotlib.pyplot.axes
        The axes handle to apply the ticker.
        By default, uses the current axes handle.

    '''
    ax, _ = __get_axes(ax=ax)

    ax.yaxis.set_major_formatter(FMT_MIDI_NOTE)
    # Get the tick labels and reset the vertical alignment
    for tick in ax.yaxis.get_ticklabels():
        tick.set_verticalalignment('baseline')


def ticker_pitch(ax=None):
    '''Set the y-axis of the given axes to MIDI frequencies

    Parameters
    ----------
    ax : matplotlib.pyplot.axes
        The axes handle to apply the ticker.
        By default, uses the current axes handle.
    '''
    ax, _ = __get_axes(ax=ax)

    ax.yaxis.set_major_formatter(FMT_MIDI_HZ)


# Instantiate ticker objects; we don't need more than one of each
FMT_MIDI_NOTE = FuncFormatter(__ticker_midi_note)
FMT_MIDI_HZ = FuncFormatter(__ticker_midi_hz)



''' Unit tests for utils
'''

import collections

import numpy as np
import nose.tools
import mir_eval
from mir_eval import util


def test_interpolate_intervals():
    """Check that an interval set is interpolated properly, with boundaries
    conditions and out-of-range values.
    """
    labels = list('abc')
    intervals = np.array([(n, n + 1.0) for n in range(len(labels))])
    time_points = [-1.0, 0.1, 0.9, 1.0, 2.3, 4.0]
    expected_ans = ['N', 'a', 'a', 'b', 'c', 'N']
    assert (util.interpolate_intervals(intervals, labels, time_points, 'N') ==
            expected_ans)


def test_interpolate_intervals_gap():
    """Check that an interval set is interpolated properly, with gaps."""
    labels = list('abc')
    intervals = np.array([[0.5, 1.0], [1.5, 2.0], [2.5, 3.0]])
    time_points = [0.0, 0.75, 1.25, 1.75, 2.25, 2.75, 3.5]
    expected_ans = ['N', 'a', 'N', 'b', 'N', 'c', 'N']
    assert (util.interpolate_intervals(intervals, labels, time_points, 'N') ==
            expected_ans)


@nose.tools.raises(ValueError)
def test_interpolate_intervals_badtime():
    """Check that interpolate_intervals throws an exception if
    input is unordered.
    """
    labels = list('abc')
    intervals = np.array([(n, n + 1.0) for n in range(len(labels))])
    time_points = [-1.0, 0.1, 0.9, 0.8, 2.3, 4.0]
    mir_eval.util.interpolate_intervals(intervals, labels, time_points)


def test_intervals_to_samples():
    """Check that an interval set is sampled properly, with boundaries
    conditions and out-of-range values.
    """
    labels = list('abc')
    intervals = np.array([(n, n + 1.0) for n in range(len(labels))])

    expected_times = [0.0, 0.5, 1.0, 1.5, 2.0, 2.5]
    expected_labels = ['a', 'a', 'b', 'b', 'c', 'c']
    result = util.intervals_to_samples(
        intervals, labels, offset=0, sample_size=0.5, fill_value='N')
    assert result[0] == expected_times
    assert result[1] == expected_labels

    expected_times = [0.25, 0.75, 1.25, 1.75, 2.25, 2.75]
    expected_labels = ['a', 'a', 'b', 'b', 'c', 'c']
    result = util.intervals_to_samples(
        intervals, labels, offset=0.25, sample_size=0.5, fill_value='N')
    assert result[0] == expected_times
    assert result[1] == expected_labels


def test_intersect_files():
    """Check that two non-identical yield correct results.
    """
    flist1 = ['/a/b/abc.lab', '/c/d/123.lab', '/e/f/xyz.lab']
    flist2 = ['/g/h/xyz.npy', '/i/j/123.txt', '/k/l/456.lab']
    sublist1, sublist2 = util.intersect_files(flist1, flist2)
    assert sublist1 == ['/e/f/xyz.lab', '/c/d/123.lab']
    assert sublist2 == ['/g/h/xyz.npy', '/i/j/123.txt']
    sublist1, sublist2 = util.intersect_files(flist1[:1], flist2[:1])
    assert sublist1 == []
    assert sublist2 == []


def test_merge_labeled_intervals():
    """Check that two labeled interval sequences merge correctly.
    """
    x_intvs = np.array([
        [0.0,    0.44],
        [0.44,  2.537],
        [2.537, 4.511],
        [4.511, 6.409]])
    x_labels = ['A', 'B', 'C', 'D']
    y_intvs = np.array([
        [0.0,   0.464],
        [0.464, 2.415],
        [2.415, 4.737],
        [4.737, 6.409]])
    y_labels = [0, 1, 2, 3]
    expected_intvs = [
        [0.0,    0.44],
        [0.44,  0.464],
        [0.464, 2.415],
        [2.415, 2.537],
        [2.537, 4.511],
        [4.511, 4.737],
        [4.737, 6.409]]
    expected_x_labels = ['A', 'B', 'B', 'B', 'C', 'D', 'D']
    expected_y_labels = [0,     0,   1,   2,   2,   2,   3]
    new_intvs, new_x_labels, new_y_labels = util.merge_labeled_intervals(
        x_intvs, x_labels, y_intvs, y_labels)

    assert new_x_labels == expected_x_labels
    assert new_y_labels == expected_y_labels
    assert new_intvs.tolist() == expected_intvs

    # Check that invalid inputs raise a ValueError
    y_intvs[-1, -1] = 10.0
    nose.tools.assert_raises(ValueError, util.merge_labeled_intervals, x_intvs,
                             x_labels, y_intvs, y_labels)


def test_boundaries_to_intervals():
    # Basic tests
    boundaries = np.arange(10)
    correct_intervals = np.array([np.arange(10 - 1), np.arange(1, 10)]).T
    intervals = mir_eval.util.boundaries_to_intervals(boundaries)
    assert np.all(intervals == correct_intervals)


def test_adjust_events():
    # Test appending at the end
    events = np.arange(1, 11)
    labels = [str(n) for n in range(10)]
    new_e, new_l = mir_eval.util.adjust_events(events, labels, 0.0, 11.)
    assert new_e[0] == 0.
    assert new_l[0] == '__T_MIN'
    assert new_e[-1] == 11.
    assert new_l[-1] == '__T_MAX'
    assert np.all(new_e[1:-1] == events)
    assert new_l[1:-1] == labels

    # Test trimming
    new_e, new_l = mir_eval.util.adjust_events(events, labels, 0.0, 9.)
    assert new_e[0] == 0.
    assert new_l[0] == '__T_MIN'
    assert new_e[-1] == 9.
    assert np.all(new_e[1:] == events[:-1])
    assert new_l[1:] == labels[:-1]


def test_bipartite_match():
    # This test constructs a graph as follows:
    #   v9 -- (u0)
    #   v8 -- (u0, u1)
    #   v7 -- (u0, u1, u2)
    #   ...
    #   v0 -- (u0, u1, ..., u9)
    #
    # This structure and ordering of this graph should force Hopcroft-Karp to
    # hit each algorithm/layering phase
    #
    G = collections.defaultdict(list)

    u_set = ['u{:d}'.format(_) for _ in range(10)]
    v_set = ['v{:d}'.format(_) for _ in range(len(u_set)+1)]
    for i, u in enumerate(u_set):
        for v in v_set[:-i-1]:
            G[v].append(u)

    matching = util._bipartite_match(G)

    # Make sure that each u vertex is matched
    nose.tools.eq_(len(matching), len(u_set))

    # Make sure that there are no duplicate keys
    lhs = set([k for k in matching])
    rhs = set([matching[k] for k in matching])

    nose.tools.eq_(len(matching), len(lhs))
    nose.tools.eq_(len(matching), len(rhs))

    # Finally, make sure that all detected edges are present in G
    for k in matching:
        v = matching[k]
        assert v in G[k] or k in G[v]


def test_outer_distance_mod_n():
    ref = [1., 2., 3.]
    est = [1.1, 6., 1.9, 5., 10.]
    expected = np.array([
        [0.1, 5., 0.9, 4., 3.],
        [0.9, 4., 0.1, 3., 4.],
        [1.9, 3., 1.1, 2., 5.]])
    actual = mir_eval.util._outer_distance_mod_n(ref, est)
    assert np.allclose(actual, expected)

    ref = [13., 14., 15.]
    est = [1.1, 6., 1.9, 5., 10.]
    expected = np.array([
        [0.1, 5., 0.9, 4., 3.],
        [0.9, 4., 0.1, 3., 4.],
        [1.9, 3., 1.1, 2., 5.]])
    actual = mir_eval.util._outer_distance_mod_n(ref, est)
    assert np.allclose(actual, expected)


def test_match_events():
    ref = [1., 2., 3.]
    est = [1.1, 6., 1.9, 5., 10.]
    expected = [(0, 0), (1, 2)]
    actual = mir_eval.util.match_events(ref, est, 0.5)
    assert actual == expected

    ref = [1., 2., 3., 11.9]
    est = [1.1, 6., 1.9, 5., 10., 0.]
    expected = [(0, 0), (1, 2), (3, 5)]
    actual = mir_eval.util.match_events(
        ref, est, 0.5, distance=mir_eval.util._outer_distance_mod_n)
    assert actual == expected


def test_fast_hit_windows():

    ref = [1., 2., 3.]
    est = [1.1, 6., 1.9, 5., 10.]

    ref_fast, est_fast = mir_eval.util._fast_hit_windows(ref, est, 0.5)
    ref_slow, est_slow = np.where(np.abs(np.subtract.outer(ref, est)) <= 0.5)

    assert np.all(ref_fast == ref_slow)
    assert np.all(est_fast == est_slow)


def test_validate_intervals():
    # Test for ValueError when interval shape is invalid
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_intervals,
        np.array([[1.], [2.5], [5.]]))
    # Test for ValueError when times are negative
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_intervals,
        np.array([[1., -2.], [2.5, 3.], [5., 6.]]))
    # Test for ValueError when duration is zero
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_intervals,
        np.array([[1., 2.], [2.5, 2.5], [5., 6.]]))
    # Test for ValueError when duration is negative
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_intervals,
        np.array([[1., 2.], [2.5, 1.5], [5., 6.]]))


def test_validate_events():
    # Test for ValueError when max_time is violated
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_events, np.array([100., 100000.]))
    # Test for ValueError when events aren't 1-d arrays
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_events,
        np.array([[1., 2.], [3., 4.]]))
    # Test for ValueError when event times are not increasing
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_events,
        np.array([1., 2., 5., 3.]))


def test_validate_frequencies():
    # Test for ValueError when max_freq is violated
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_frequencies,
        np.array([100., 100000.]), 5000., 20.)
    # Test for ValueError when min_freq is violated
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_frequencies,
        np.array([2., 200.]), 5000., 20.)
    # Test for ValueError when events aren't 1-d arrays
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_frequencies,
        np.array([[100., 200.], [300., 400.]]), 5000., 20.)
    # Test for ValueError when allow_negatives is false and negative values
    # are passed
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_frequencies,
        np.array([[-100., 200.], [300., 400.]]), 5000., 20.,
        allow_negatives=False)
    # Test for ValueError when max_freq is violated and allow_negatives=True
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_frequencies,
        np.array([100., -100000.]), 5000., 20., allow_negatives=True)
    # Test for ValueError when min_freq is violated and allow_negatives=True
    nose.tools.assert_raises(
        ValueError, mir_eval.util.validate_frequencies,
        np.array([-2., 200.]), 5000., 20., allow_negatives=True)


def test_has_kwargs():

    def __test(target, f):
        assert target == mir_eval.util.has_kwargs(f)

    def f1(_):
        return None

    def f2(_=5):
        return None

    def f3(*_):
        return None

    def f4(_, **kw):
        return None

    def f5(_=5, **kw):
        return None

    yield __test, False, f1
    yield __test, False, f2
    yield __test, False, f3
    yield __test, True, f4
    yield __test, True, f5


def test_sort_labeled_intervals():

    def __test_labeled(x, labels, x_true, lab_true):
        xs, ls = mir_eval.util.sort_labeled_intervals(x, labels)

        assert np.allclose(xs, x_true)
        nose.tools.eq_(ls, lab_true)

    def __test(x, x_true):
        xs = mir_eval.util.sort_labeled_intervals(x)
        assert np.allclose(xs, x_true)

    x1 = np.asarray([[10, 20], [0, 10]])
    x1_true = np.asarray([[0, 10], [10, 20]])
    labels = ['a', 'b']
    labels_true = ['b', 'a']

    yield __test_labeled, x1, labels, x1_true, labels_true
    yield __test, x1, x1_true
    yield __test_labeled, x1_true, labels_true, x1_true, labels_true
    yield __test, x1_true, x1_true



"""
Alignment models are given a sequence of events along with a piece of audio, and then return a
sequence of timestamps, with one timestamp for each event, indicating the position of this event
in the audio. The events are listed in order of occurrence in the audio, so that output
timestamps have to be monotonically increasing.
Evaluation usually involves taking the series of predicted and ground truth timestamps and
comparing their distance, usually on a pair-wise basis, e.g. taking the median absolute error in
seconds.

Conventions
-----------
Timestamps should be provided in the form of a 1-dimensional array of onset
times in seconds in increasing order.

Metrics
-------
* :func:`mir_eval.alignment.absolute_error`: Median absolute error and average absolute error
* :func:`mir_eval.alignment.percentage_correct`: Percentage of correct timestamps,
where a timestamp is counted
as correct if it lies within a certain tolerance window around the ground truth timestamp
* :func:`mir_eval.alignment.pcs`: Percentage of correct segments: Percentage of overlap between
predicted segments and ground truth segments, where segments are defined by (start time,
end time) pairs
* :func:`mir_eval.alignment.perceptual_metric`: metric based on human synchronicity perception as
measured in the paper "User-centered evaluation of lyrics to audio alignment",
N. Lizé-Masclef, A. Vaglio, M. Moussallam, ISMIR 2021

References
----------
  .. [#lizemasclef2021] N. Lizé-Masclef, A. Vaglio, M. Moussallam.
    "User-centered evaluation of lyrics to audio alignment",
    International Society for Music Information Retrieval (ISMIR) conference,
    2021.

  .. [#mauch2010] M. Mauch, F: Hiromasa, M. Goto.
    "Lyrics-to-audio alignment and phrase-level segmentation using
    incomplete internet-style chord annotations",
    Frontiers in Proceedings of the Sound Music Computing Conference (SMC), 2010.

  .. [#dzhambazov2017] G. Dzhambazov.
    "Knowledge-Based Probabilistic Modeling For Tracking Lyrics In Music Audio Signals",
    PhD Thesis, 2017.

  .. [#fujihara2011] H. Fujihara, M. Goto, J. Ogata, H. Okuno.
    "LyricSynchronizer: Automatic synchronization system between musical audio signals and lyrics",
    IEEE Journal of Selected Topics in Signal Processing, VOL. 5, NO. 6, 2011

"""

import collections
from typing import Optional

import numpy as np
from scipy.stats import skewnorm

from mir_eval.util import filter_kwargs


def validate(
    reference_timestamps: np.ndarray, estimated_timestamps: np.ndarray
):
    """Checks that the input annotations to a metric look like valid onset time
    arrays, and throws helpful errors if not.

    Parameters
    ----------
    reference_timestamps : np.ndarray
        reference timestamp locations, in seconds
    estimated_timestamps : np.ndarray
        estimated timestamp locations, in seconds
    """
    # We need to have 1D numpy arrays
    if not isinstance(reference_timestamps, np.ndarray):
        raise ValueError(
            "Reference timestamps need to be a numpy array, but got"
            f" {type(reference_timestamps)}"
        )
    if not isinstance(estimated_timestamps, np.ndarray):
        raise ValueError(
            "Estimated timestamps need to be a numpy array, but got"
            f" {type(estimated_timestamps)}"
        )
    if reference_timestamps.ndim != 1:
        raise ValueError(
            "Reference timestamps need to be a one-dimensional vector, but got"
            f" {reference_timestamps.ndim} dimensions"
        )
    if estimated_timestamps.ndim != 1:
        raise ValueError(
            "Estimated timestamps need to be a one-dimensional vector, but got"
            f" {estimated_timestamps.ndim} dimensions"
        )

    # If reference or estimated timestamps are empty, cannot compute metric
    if reference_timestamps.size == 0:
        raise ValueError("Reference timestamps are empty.")
    if estimated_timestamps.size != reference_timestamps.size:
        raise ValueError(
            "Number of timestamps must be the same in prediction and ground"
            f" truth, but found {estimated_timestamps.size} in prediction and"
            f" {reference_timestamps.size} in ground truth"
        )

    # Check monotonicity
    if not np.all(reference_timestamps[1:] - reference_timestamps[:-1] >= 0):
        raise ValueError(
            "Reference timestamps are not monotonically increasing!"
        )
    if not np.all(estimated_timestamps[1:] - estimated_timestamps[:-1] >= 0):
        raise ValueError(
            "Estimated timestamps are not monotonically increasing!"
        )

    # Check positivity (need for correct PCS metric calculation)
    if not np.all(reference_timestamps >= 0):
        raise ValueError("Reference timestamps can not be below 0!")
    if not np.all(estimated_timestamps >= 0):
        raise ValueError("Estimated timestamps can not be below 0!")


def absolute_error(reference_timestamps, estimated_timestamps):
    """Compute the absolute deviations between estimated and reference timestamps,
    and then returns the median and average over all events

    Examples
    --------
    >>> reference_timestamps = mir_eval.io.load_events('reference.txt')
    >>> estimated_timestamps = mir_eval.io.load_events('estimated.txt')
    >>> mae, aae = mir_eval.align.absolute_error(reference_onsets, estimated_timestamps)

    Parameters
    ----------
    reference_timestamps : np.ndarray
        reference timestamps, in seconds
    estimated_timestamps : np.ndarray
        estimated timestamps, in seconds

    Returns
    -------
    mae : float
        Median absolute error
    aae: float
        Average absolute error
    """
    validate(reference_timestamps, estimated_timestamps)
    deviations = np.abs(reference_timestamps - estimated_timestamps)
    return np.median(deviations), np.mean(deviations)


def percentage_correct(reference_timestamps, estimated_timestamps, window=0.3):
    """Compute the percentage of correctly predicted timestamps. A timestamp is predicted
    correctly if its position doesn't deviate more than the window parameter from the ground
    truth timestamp.

    Examples
    --------
    >>> reference_timestamps = mir_eval.io.load_events('reference.txt')
    >>> estimated_timestamps = mir_eval.io.load_events('estimated.txt')
    >>> pc = mir_eval.align.percentage_correct(reference_onsets, estimated_timestamps, window=0.2)

    Parameters
    ----------
    reference_timestamps : np.ndarray
        reference timestamps, in seconds
    estimated_timestamps : np.ndarray
        estimated timestamps, in seconds
    window : float
        Window size, in seconds
        (Default value = .3)

    Returns
    -------
    pc : float
        Percentage of correct timestamps
    """
    validate(reference_timestamps, estimated_timestamps)
    deviations = np.abs(reference_timestamps - estimated_timestamps)
    return np.mean(deviations <= window)


def percentage_correct_segments(
    reference_timestamps, estimated_timestamps, duration: Optional[float] = None
):
    """Calculates the percentage of correct segments (PCS) metric.

    It constructs segments out of predicted and estimated timestamps separately
    out of each given timestamp vector and calculates the percentage of overlap between correct
    segments compared to the total duration.

    WARNING: This metrics behaves differently depending on whether "duration" is given!

    If duration is not given (default case), the computation follows the MIREX lyrics alignment
    challenge 2020. For a timestamp vector with entries (t1,t2, ... tN), segments with
    the following (start, end) boundaries are created: (t1, t2), ... (tN-1, tN).
    After the segments are created, the overlap between the reference and estimated segments is
    determined and divided by the total duration, which is the distance between the
    first and last timestamp in the reference.

    If duration is given, the segment boundaries are instead (0, t1), (t1, t2), ... (tN, duration).
    The overlap is computed in the same way, but then divided by the duration parameter given to
    this function.
    This method follows the original paper [#fujihara2011] more closely, where the metric was
    proposed.
    As a result, this variant of the metrics punishes cases where the first estimated timestamp
    is too early or the last estimated timestamp is too late, whereas the MIREX variant does not.
    On the other hand, the MIREX metric is invariant to how long the eventless beginning and end
    parts of the audio are, which might be a desirable property.

    Examples
    --------
    >>> reference_timestamps = mir_eval.io.load_events('reference.txt')
    >>> estimated_timestamps = mir_eval.io.load_events('estimated.txt')
    >>> pcs = mir_eval.align.percentage_correct_segments(reference_timestamps, estimated_timestamps)

    Parameters
    ----------
    reference_timestamps : np.ndarray
        reference timestamps, in seconds
    estimated_timestamps : np.ndarray
        estimated timestamps, in seconds
    duration : float
        Optional. Total duration of audio (seconds). WARNING: Metric is computed differently
        depending on whether this is provided or not - see documentation above!

    Returns
    -------
    pcs : float
        Percentage of time where ground truth and predicted segments overlap
    """
    validate(reference_timestamps, estimated_timestamps)
    if duration is not None:
        duration = float(duration)
        if duration <= 0:
            raise ValueError(
                f"Positive duration needs to be provided, but got {duration}"
            )
        if np.max(reference_timestamps) > duration:
            raise ValueError(
                "Expected largest reference timestamp"
                f"{np.max(reference_timestamps)} to not be "
                f"larger than duration {duration}"
            )
        if np.max(estimated_timestamps) > duration:
            raise ValueError(
                "Expected largest estimated timestamp "
                f"{np.max(estimated_timestamps)} to not be "
                f"larger than duration {duration}"
            )

        ref_starts = np.concatenate([[0], reference_timestamps])
        ref_ends = np.concatenate([reference_timestamps, [duration]])
        est_starts = np.concatenate([[0], estimated_timestamps])
        est_ends = np.concatenate([estimated_timestamps, [duration]])
    else:
        # MIREX lyrics alignment 2020 style:
        # Ignore regions before start and after end reference timestamp
        duration = reference_timestamps[-1] - reference_timestamps[0]
        if duration <= 0:
            raise ValueError(
                f"Reference timestamps are all identical, can not compute PCS"
                f" metric!"
            )

        ref_starts = reference_timestamps[:-1]
        ref_ends = reference_timestamps[1:]
        est_starts = estimated_timestamps[:-1]
        est_ends = estimated_timestamps[1:]

    overlap_starts = np.maximum(ref_starts, est_starts)
    overlap_ends = np.minimum(ref_ends, est_ends)
    overlap_duration = np.sum(np.maximum(overlap_ends - overlap_starts, 0))
    return overlap_duration / duration


def karaoke_perceptual_metric(reference_timestamps, estimated_timestamps):
    """Metric based on human synchronicity perception as measured in the paper
    "User-centered evaluation of lyrics to audio alignment" [#lizemasclef2021]

    The parameters of this function were tuned on data collected through a user Karaoke-like
    experiment
    It reflects human judgment of how "synchronous" lyrics and audio stimuli are perceived
    in that setup.
    Beware that this metric is non-symmetrical and by construction it is also not equal to 1 at 0.

    Examples
    --------
    >>> reference_timestamps = mir_eval.io.load_events('reference.txt')
    >>> estimated_timestamps = mir_eval.io.load_events('estimated.txt')
    >>> score = mir_eval.align.karaoke_perceptual_metric(reference_onsets, estimated_timestamps)

    Parameters
    ----------
    reference_timestamps : np.ndarray
        reference timestamps, in seconds
    estimated_timestamps : np.ndarray
        estimated timestamps, in seconds

    Returns
    -------
    perceptual_score : float
        Perceptual score, averaged over all timestamps
    """
    validate(reference_timestamps, estimated_timestamps)
    offsets = estimated_timestamps - reference_timestamps

    # Score offsets using a certain skewed normal distribution
    skewness = 1.12244251
    localisation = -0.22270315
    scale = 0.29779424
    normalisation_factor = 1.6857
    perceptual_scores = (1.0 / normalisation_factor) * skewnorm.pdf(
        offsets, skewness, loc=localisation, scale=scale
    )

    return np.mean(perceptual_scores)


def evaluate(reference_timestamps, estimated_timestamps, **kwargs):
    """Compute all metrics for the given reference and estimated annotations.
    Examples
    --------
    >>> reference_timestamps = mir_eval.io.load_events('reference.txt')
    >>> estimated_timestamps = mir_eval.io.load_events('estimated.txt')
    >>> duration = max(np.max(reference_timestamps), np.max(estimated_timestamps)) + 10
    >>> scores = mir_eval.align.evaluate(reference_onsets, estimated_timestamps, duration)

    Parameters
    ----------
    reference_timestamps : np.ndarray
        reference timestamp locations, in seconds
    estimated_timestamps : np.ndarray
        estimated timestamp locations, in seconds
    kwargs
        Additional keyword arguments which will be passed to the
        appropriate metric or preprocessing functions.

    Returns
    -------
    scores : dict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.
    """
    # Compute all metrics
    scores = collections.OrderedDict()

    scores["pc"] = filter_kwargs(
        percentage_correct, reference_timestamps, estimated_timestamps, **kwargs
    )
    scores["mae"], scores["aae"] = absolute_error(
        reference_timestamps, estimated_timestamps
    )
    scores["pcs"] = filter_kwargs(
        percentage_correct_segments,
        reference_timestamps,
        estimated_timestamps,
        **kwargs,
    )
    scores["perceptual"] = karaoke_perceptual_metric(
        reference_timestamps, estimated_timestamps
    )

    return scores



'''
Tests for mir_eval.key
'''

import mir_eval
import nose.tools
import glob
import json
import numpy as np

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/key/ref*.txt'
EST_GLOB = 'data/key/est*.txt'
SCORES_GLOB = 'data/key/output*.json'


def __unit_test_key_function(metric):

    good_keys = ['C major', 'c major', 'C# major', 'Bb minor',
                 'db minor', 'X', 'x', 'C other']
    # All of these are invalid key strings
    bad_keys = ['C maj', 'Cb major', 'C', 'K major', 'F## minor'
                'X other', 'x minor']

    for good_key in good_keys:
        for bad_key in bad_keys:
            # Should raise an error whether we pass a bad key as ref or est
            nose.tools.assert_raises(ValueError, metric, good_key, bad_key)
            nose.tools.assert_raises(ValueError, metric, bad_key, good_key)

    for good_key in good_keys:
        # When the same key is passed for est and ref, score should be 1
        assert mir_eval.key.weighted_score(good_key, good_key) == 1.


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_key_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit test all metrics (one for now)
    for metric in [mir_eval.key.weighted_score]:
        yield __unit_test_key_function, metric

    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in an example key annotation
        reference_key = mir_eval.io.load_key(ref_f)
        # Load in an example key detector output
        estimated_key = mir_eval.io.load_key(est_f)
        # Compute scores
        scores = mir_eval.key.evaluate(reference_key, estimated_key)
        # Compare them
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])



"""
Some unit tests for the pattern discovery task.
"""

import numpy as np
import json
import mir_eval
import glob
import warnings
import nose.tools

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/pattern/ref*.txt'
EST_GLOB = 'data/pattern/est*.txt'
SCORES_GLOB = 'data/pattern/output*.json'


def __unit_test_pattern_function(metric):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for a warning on empty pattern
        metric([[[]]], [[[(100, 20)]]])
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == 'Reference patterns are empty.'
        metric([[[(100, 20)]]], [[[]]])
        assert len(w) == 2
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Estimated patterns are empty."
        # And that the metric is 0
        assert np.allclose(metric([[[]]], [[[]]]), 0)

    # Now test validation function - patterns must contain at least 1 occ
    patterns = [[[(100, 20)]], []]
    nose.tools.assert_raises(ValueError, metric, patterns, patterns)
    # The (onset, midi) tuple must contain 2 elements
    patterns = [[[(100, 20, 3)]]]
    nose.tools.assert_raises(ValueError, metric, patterns, patterns)

    # Valid patterns which are the same produce a score of 1 for all metrics
    patterns = [[[(100, 20), (200, 30)]]]
    assert np.allclose(metric(patterns, patterns), 1)


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_pattern_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit tests
    for metric in [mir_eval.pattern.standard_FPR,
                   mir_eval.pattern.establishment_FPR,
                   mir_eval.pattern.occurrence_FPR,
                   mir_eval.pattern.three_layer_FPR,
                   mir_eval.pattern.first_n_three_layer_P,
                   mir_eval.pattern.first_n_target_proportion_R]:
        yield (__unit_test_pattern_function, metric)
    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in reference and estimated patterns
        reference_patterns = mir_eval.io.load_patterns(ref_f)
        estimated_patterns = mir_eval.io.load_patterns(est_f)
        # Compute scores
        scores = mir_eval.pattern.evaluate(reference_patterns,
                                           estimated_patterns)
        # Compare them
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])



# CREATED: 2/9/16 2:27 PM by Justin Salamon <justin.salamon@nyu.edu>

import mir_eval
import numpy as np
import glob
import json
from nose.tools import raises
import warnings

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/transcription/ref*.txt'
EST_GLOB = 'data/transcription/est*.txt'
SCORES_GLOB = 'data/transcription/output*.json'

REF = np.array([
    [0.100, 0.300, 220.000],
    [0.300, 0.400, 246.942],
    [0.500, 0.600, 277.183],
    [0.550, 0.650, 293.665]])

EST = np.array([
    [0.120,   0.290,   225.000],
    [0.300,   0.340,   246.942],
    [0.500,   0.600,   500.000],
    [0.550,   0.600,   293.665],
    [0.560,   0.650,   293.665]])

SCORES = {
    "Precision": 0.4,
    "Recall": 0.5,
    "F-measure": 0.4444444444444445,
    "Average_Overlap_Ratio": 0.675,
    "Precision_no_offset": 0.6,
    "Recall_no_offset": 0.75,
    "F-measure_no_offset": 0.6666666666666665,
    "Average_Overlap_Ratio_no_offset": 0.5833333333333333
}

ONSET_SCORES = {
    "Onset_Precision": 0.8,
    "Onset_Recall": 1.0,
    "Onset_F-measure": 0.8888888888888889,
}

OFFSET_SCORES = {
    "Offset_Precision": 0.6,
    "Offset_Recall": 0.75,
    "Offset_F-measure": 0.6666666666666665,
}


def test_match_note_offsets():

    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = (
        mir_eval.transcription.match_note_offsets(ref_int, est_int))

    assert matching == [(0, 0), (2, 2), (3, 3)]


def test_match_note_offsets_strict():

    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = (
        mir_eval.transcription.match_note_offsets(
            ref_int, est_int, strict=True))

    assert matching == [(0, 0), (2, 2), (3, 4)]


def test_match_note_onsets():

    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = (
        mir_eval.transcription.match_note_onsets(ref_int, est_int))

    assert matching == [(0, 0), (1, 1), (2, 2), (3, 3)]


def test_match_note_onsets_strict():

    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = (
        mir_eval.transcription.match_note_onsets(
            ref_int, est_int, strict=True))

    assert matching == [(0, 0), (1, 1), (2, 2), (3, 3)]


def test_match_notes():

    ref_int, ref_pitch = REF[:, :2], REF[:, 2]
    est_int, est_pitch = EST[:, :2], EST[:, 2]

    matching = (
        mir_eval.transcription.match_notes(ref_int, ref_pitch, est_int,
                                           est_pitch))

    assert matching == [(0, 0), (3, 3)]

    matching = (
        mir_eval.transcription.match_notes(ref_int, ref_pitch, est_int,
                                           est_pitch, offset_ratio=None))

    assert matching == [(0, 0), (1, 1), (3, 3)]


def test_match_notes_strict():

    ref_int, ref_pitch = np.array([[0, 1]]), np.array([100])
    est_int, est_pitch = np.array([[0.05, 1]]), np.array([100])

    matching = (
        mir_eval.transcription.match_notes(ref_int, ref_pitch, est_int,
                                           est_pitch, strict=True))

    assert matching == []


def test_precision_recall_f1_overlap():

    # load test data
    ref_int, ref_pitch = REF[:, :2], REF[:, 2]
    est_int, est_pitch = EST[:, :2], EST[:, 2]

    precision, recall, f_measure, avg_overlap_ratio = (
        mir_eval.transcription.precision_recall_f1_overlap(
            ref_int, ref_pitch, est_int, est_pitch))

    scores_gen = np.array([precision, recall, f_measure, avg_overlap_ratio])
    scores_exp = np.array([SCORES['Precision'], SCORES['Recall'],
                           SCORES['F-measure'],
                           SCORES['Average_Overlap_Ratio']])
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)

    precision, recall, f_measure, avg_overlap_ratio = (
        mir_eval.transcription.precision_recall_f1_overlap(
            ref_int, ref_pitch, est_int, est_pitch, offset_ratio=None))

    scores_gen = np.array([precision, recall, f_measure, avg_overlap_ratio])
    scores_exp = np.array([SCORES['Precision_no_offset'],
                           SCORES['Recall_no_offset'],
                           SCORES['F-measure_no_offset'],
                           SCORES['Average_Overlap_Ratio_no_offset']])
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)


def __check_score(score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_onset_precision_recall_f1():

    # load test data
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    precision, recall, f_measure = (
        mir_eval.transcription.onset_precision_recall_f1(ref_int, est_int))

    scores_gen = np.array([precision, recall, f_measure])
    scores_exp = np.array([ONSET_SCORES['Onset_Precision'],
                           ONSET_SCORES['Onset_Recall'],
                           ONSET_SCORES['Onset_F-measure']])
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)


def test_offset_precision_recall_f1():

    # load test data
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    precision, recall, f_measure = (
        mir_eval.transcription.offset_precision_recall_f1(ref_int, est_int))

    scores_gen = np.array([precision, recall, f_measure])
    scores_exp = np.array([OFFSET_SCORES['Offset_Precision'],
                           OFFSET_SCORES['Offset_Recall'],
                           OFFSET_SCORES['Offset_F-measure']])
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)


def test_regression():

    # Regression tests
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in reference transcription
        ref_int, ref_pitch = mir_eval.io.load_valued_intervals(ref_f)
        # Load in estimated transcription
        est_int, est_pitch = mir_eval.io.load_valued_intervals(est_f)
        scores = mir_eval.transcription.evaluate(ref_int, ref_pitch, est_int,
                                                 est_pitch)
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, scores[metric], expected_scores[metric])


def test_invalid_pitch():

    ref_int, ref_pitch = np.array([[0, 1]]), np.array([-100])
    est_int, est_pitch = np.array([[0, 1]]), np.array([100])

    yield (raises(ValueError)(mir_eval.transcription.validate),
           ref_int, ref_pitch, est_int, est_pitch)
    yield (raises(ValueError)(mir_eval.transcription.validate),
           est_int, est_pitch, ref_int, ref_pitch)


def test_inconsistent_int_pitch():

    ref_int, ref_pitch = np.array([[0, 1], [2, 3]]), np.array([100])
    est_int, est_pitch = np.array([[0, 1]]), np.array([100])

    yield (raises(ValueError)(mir_eval.transcription.validate),
           ref_int, ref_pitch, est_int, est_pitch)
    yield (raises(ValueError)(mir_eval.transcription.validate),
           est_int, est_pitch, ref_int, ref_pitch)


def test_empty_ref():

    warnings.resetwarnings()
    warnings.simplefilter('always')
    with warnings.catch_warnings(record=True) as out:

        ref_int, ref_pitch = np.empty(shape=(0, 2)), np.array([])
        est_int, est_pitch = np.array([[0, 1]]), np.array([100])

        mir_eval.transcription.validate(ref_int, ref_pitch, est_int, est_pitch)

        # Make sure that the warning triggered
        assert len(out) > 0

        # And that the category is correct
        assert out[0].category is UserWarning

        # And that it says the right thing (roughly)
        assert 'empty' in str(out[0].message).lower()


def test_empty_est():

    warnings.resetwarnings()
    warnings.simplefilter('always')
    with warnings.catch_warnings(record=True) as out:

        ref_int, ref_pitch = np.array([[0, 1]]), np.array([100])
        est_int, est_pitch = np.empty(shape=(0, 2)), np.array([])

        mir_eval.transcription.validate(ref_int, ref_pitch, est_int, est_pitch)

        # Make sure that the warning triggered
        assert len(out) > 0

        # And that the category is correct
        assert out[0].category is UserWarning

        # And that it says the right thing (roughly)
        assert 'empty' in str(out[0].message).lower()


def test_precision_recall_f1_overlap_empty():

    ref_int, ref_pitch = np.empty(shape=(0, 2)), np.array([])
    est_int, est_pitch = np.array([[0, 1]]), np.array([100])

    precision, recall, f1, avg_overlap_ratio = (
        mir_eval.transcription.precision_recall_f1_overlap(
            ref_int, ref_pitch, est_int, est_pitch))

    assert (precision, recall, f1) == (0, 0, 0)

    precision, recall, f1, avg_overlap_ratio = (
        mir_eval.transcription.precision_recall_f1_overlap(
            est_int, est_pitch, ref_int, ref_pitch))

    assert (precision, recall, f1) == (0, 0, 0)


def test_onset_precision_recall_f1_empty():

    ref_int = np.empty(shape=(0, 2))
    est_int = np.array([[0, 1]])

    precision, recall, f1 = (
        mir_eval.transcription.onset_precision_recall_f1(ref_int, est_int))

    assert (precision, recall, f1) == (0, 0, 0)

    precision, recall, f1 = (
        mir_eval.transcription.onset_precision_recall_f1(est_int, ref_int))

    assert (precision, recall, f1) == (0, 0, 0)


def test_offset_precision_recall_f1_empty():

    ref_int = np.empty(shape=(0, 2))
    est_int = np.array([[0, 1]])

    precision, recall, f1 = (
        mir_eval.transcription.offset_precision_recall_f1(ref_int, est_int))

    assert (precision, recall, f1) == (0, 0, 0)

    precision, recall, f1 = (
        mir_eval.transcription.offset_precision_recall_f1(est_int, ref_int))

    assert (precision, recall, f1) == (0, 0, 0)



# CREATED:2015-09-16 14:46:47 by Brian McFee <brian.mcfee@nyu.edu>
# -*- encoding: utf-8 -*-
'''Evaluation criteria for hierarchical structure analysis.

Hierarchical structure analysis seeks to annotate a track with a nested
decomposition of the temporal elements of the piece, effectively providing
a kind of "parse tree" of the composition.  Unlike the flat segmentation
metrics defined in :mod:`mir_eval.segment`, which can only encode one level of
analysis, hierarchical annotations expose the relationships between short
segments and the larger compositional elements to which they belong.

Conventions
-----------
Annotations are assumed to take the form of an ordered list of segmentations.
As in the :mod:`mir_eval.segment` metrics, each segmentation itself consists of
an n-by-2 array of interval times, so that the ``i`` th segment spans time
``intervals[i, 0]`` to ``intervals[i, 1]``.

Hierarchical annotations are ordered by increasing specificity, so that the
first segmentation should contain the fewest segments, and the last
segmentation contains the most.

Metrics
-------
* :func:`mir_eval.hierarchy.tmeasure`: Precision, recall, and F-measure of
  triplet-based frame accuracy for boundary detection.

* :func:`mir_eval.hierarchy.lmeasure`: Precision, recall, and F-measure of
  triplet-based frame accuracy for segment labeling.

References
----------
  .. [#mcfee2015] Brian McFee, Oriol Nieto, and Juan P. Bello.
    "Hierarchical evaluation of segment boundary detection",
    International Society for Music Information Retrieval (ISMIR) conference,
    2015.

  .. [#mcfee2017] Brian McFee, Oriol Nieto, Morwaread Farbood, and
    Juan P. Bello.
    "Evaluating hierarchical structure in music annotations",
    Frontiers in Psychology, 2017.
'''

import collections
import itertools
import warnings

import numpy as np
import scipy.sparse

from . import util
from .segment import validate_structure


def _round(t, frame_size):
    '''Round a time-stamp to a specified resolution.

    Equivalent to ``t - np.mod(t, frame_size)``.

    Examples
    --------
    >>> _round(53.279, 0.1)
    53.2
    >>> _round(53.279, 0.25)
    53.25

    Parameters
    ----------
    t : number or ndarray
        The time-stamp to round

    frame_size : number > 0
        The resolution to round to

    Returns
    -------
    t_round : number
        The rounded time-stamp
    '''
    return t - np.mod(t, float(frame_size))


def _hierarchy_bounds(intervals_hier):
    '''Compute the covered time range of a hierarchical segmentation.

    Parameters
    ----------
    intervals_hier : list of ndarray
        A hierarchical segmentation, encoded as a list of arrays of segment
        intervals.

    Returns
    -------
    t_min : float
    t_max : float
        The minimum and maximum times spanned by the annotation
    '''
    boundaries = list(itertools.chain(*list(itertools.chain(*intervals_hier))))

    return min(boundaries), max(boundaries)


def _align_intervals(int_hier, lab_hier, t_min=0.0, t_max=None):
    '''Align a hierarchical annotation to span a fixed start and end time.

    Parameters
    ----------
    int_hier : list of list of intervals
    lab_hier : list of list of str
        Hierarchical segment annotations, encoded as a
        list of list of intervals (int_hier) and list of
        list of strings (lab_hier)

    t_min : None or number >= 0
        The minimum time value for the segmentation

    t_max : None or number >= t_min
        The maximum time value for the segmentation

    Returns
    -------
    intervals_hier : list of list of intervals
    labels_hier : list of list of str
        `int_hier` `lab_hier` aligned to span `[t_min, t_max]`.
    '''
    return [list(_) for _ in zip(*[util.adjust_intervals(np.asarray(ival),
                                                         labels=lab,
                                                         t_min=t_min,
                                                         t_max=t_max)
                                   for ival, lab in zip(int_hier, lab_hier)])]


def _lca(intervals_hier, frame_size):
    '''Compute the (sparse) least-common-ancestor (LCA) matrix for a
    hierarchical segmentation.

    For any pair of frames ``(s, t)``, the LCA is the deepest level in
    the hierarchy such that ``(s, t)`` are contained within a single
    segment at that level.

    Parameters
    ----------
    intervals_hier : list of ndarray
        An ordered list of segment interval arrays.
        The list is assumed to be ordered by increasing specificity (depth).

    frame_size : number
        The length of the sample frames (in seconds)

    Returns
    -------
    lca_matrix : scipy.sparse.csr_matrix
        A sparse matrix such that ``lca_matrix[i, j]`` contains the depth
        of the deepest segment containing frames ``i`` and ``j``.
    '''

    frame_size = float(frame_size)

    # Figure out how many frames we need

    n_start, n_end = _hierarchy_bounds(intervals_hier)

    n = int((_round(n_end, frame_size) -
             _round(n_start, frame_size)) / frame_size)

    # Initialize the LCA matrix
    lca_matrix = scipy.sparse.lil_matrix((n, n), dtype=np.uint8)

    for level, intervals in enumerate(intervals_hier, 1):
        for ival in (_round(np.asarray(intervals),
                            frame_size) / frame_size).astype(int):
            idx = slice(ival[0], ival[1])
            lca_matrix[idx, idx] = level

    return lca_matrix.tocsr()


def _meet(intervals_hier, labels_hier, frame_size):
    '''Compute the (sparse) least-common-ancestor (LCA) matrix for a
    hierarchical segmentation.

    For any pair of frames ``(s, t)``, the LCA is the deepest level in
    the hierarchy such that ``(s, t)`` are contained within a single
    segment at that level.

    Parameters
    ----------
    intervals_hier : list of ndarray
        An ordered list of segment interval arrays.
        The list is assumed to be ordered by increasing specificity (depth).

    labels_hier : list of list of str
        ``labels_hier[i]`` contains the segment labels for the
        ``i``th layer of the annotations

    frame_size : number
        The length of the sample frames (in seconds)

    Returns
    -------
    meet_matrix : scipy.sparse.csr_matrix
        A sparse matrix such that ``meet_matrix[i, j]`` contains the depth
        of the deepest segment label containing both ``i`` and ``j``.
    '''

    frame_size = float(frame_size)

    # Figure out how many frames we need

    n_start, n_end = _hierarchy_bounds(intervals_hier)

    n = int((_round(n_end, frame_size) -
             _round(n_start, frame_size)) / frame_size)

    # Initialize the meet matrix
    meet_matrix = scipy.sparse.lil_matrix((n, n), dtype=np.uint8)

    for level, (intervals, labels) in enumerate(zip(intervals_hier,
                                                    labels_hier), 1):

        # Encode the labels at this level
        lab_enc = util.index_labels(labels)[0]

        # Find unique agreements
        int_agree = np.triu(np.equal.outer(lab_enc, lab_enc))

        # Map intervals to frame indices
        int_frames = (_round(intervals, frame_size) / frame_size).astype(int)

        # For each intervals i, j where labels agree, update the meet matrix
        for (seg_i, seg_j) in zip(*np.where(int_agree)):
            idx_i = slice(*list(int_frames[seg_i]))
            idx_j = slice(*list(int_frames[seg_j]))
            meet_matrix[idx_i, idx_j] = level
            if seg_i != seg_j:
                meet_matrix[idx_j, idx_i] = level

    return scipy.sparse.csr_matrix(meet_matrix)


def _gauc(ref_lca, est_lca, transitive, window):
    '''Generalized area under the curve (GAUC)

    This function computes the normalized recall score for correctly
    ordering triples ``(q, i, j)`` where frames ``(q, i)`` are closer than
    ``(q, j)`` in the reference annotation.

    Parameters
    ----------
    ref_lca : scipy.sparse
    est_lca : scipy.sparse
        The least common ancestor matrices for the reference and
        estimated annotations

    transitive : bool
        If True, then transitive comparisons are counted, meaning that
        ``(q, i)`` and ``(q, j)`` can differ by any number of levels.

        If False, then ``(q, i)`` and ``(q, j)`` can differ by exactly one
        level.

    window : number or None
        The maximum number of frames to consider for each query.
        If `None`, then all frames are considered.

    Returns
    -------
    score : number [0, 1]
        The percentage of reference triples correctly ordered by
        the estimation.

    Raises
    ------
    ValueError
        If ``ref_lca`` and ``est_lca`` have different shapes
    '''
    # Make sure we have the right number of frames

    if ref_lca.shape != est_lca.shape:
        raise ValueError('Estimated and reference hierarchies '
                         'must have the same shape.')

    # How many frames?
    n = ref_lca.shape[0]

    # By default, the window covers the entire track
    if window is None:
        window = n

    # Initialize the score
    score = 0.0

    # Iterate over query frames
    num_frames = 0

    for query in range(n):

        # Find all pairs i,j such that ref_lca[q, i] > ref_lca[q, j]
        results = slice(max(0, query - window), min(n, query + window))

        ref_score = ref_lca[query, results]
        est_score = est_lca[query, results]

        # Densify the results
        ref_score = ref_score.toarray().squeeze()
        est_score = est_score.toarray().squeeze()

        # Don't count the query as a result
        # when query < window, query itself is the index within the slice
        # otherwise, query is located at the center of the slice, window
        # (this also holds when the slice goes off the end of the array.)
        idx = min(query, window)

        ref_score = np.concatenate((ref_score[:idx], ref_score[idx+1:]))
        est_score = np.concatenate((est_score[:idx], est_score[idx+1:]))

        inversions, normalizer = _compare_frame_rankings(ref_score, est_score,
                                                         transitive=transitive)

        if normalizer:
            score += 1.0 - inversions / float(normalizer)
            num_frames += 1

    # Normalize by the number of frames counted.
    # If no frames are counted, take the convention 0/0 -> 0
    if num_frames:
        score /= float(num_frames)
    else:
        score = 0.0

    return score


def _count_inversions(a, b):
    '''Count the number of inversions in two numpy arrays:

    # points i, j where a[i] >= b[j]

    Parameters
    ----------
    a, b : np.ndarray, shape=(n,) (m,)
        The arrays to be compared.

        This implementation is optimized for arrays with many
        repeated values.

    Returns
    -------
    inversions : int
        The number of detected inversions
    '''

    a, a_counts = np.unique(a, return_counts=True)
    b, b_counts = np.unique(b, return_counts=True)

    inversions = 0
    i = 0
    j = 0

    while i < len(a) and j < len(b):
        if a[i] < b[j]:
            i += 1
        elif a[i] >= b[j]:
            inversions += np.sum(a_counts[i:]) * b_counts[j]
            j += 1

    return inversions


def _compare_frame_rankings(ref, est, transitive=False):
    '''Compute the number of ranking disagreements in two lists.

    Parameters
    ----------
    ref : np.ndarray, shape=(n,)
    est : np.ndarray, shape=(n,)
        Reference and estimate ranked lists.
        `ref[i]` is the relevance score for point `i`.

    transitive : bool
        If true, all pairs of reference levels are compared.
        If false, only adjacent pairs of reference levels are compared.

    Returns
    -------
    inversions : int
        The number of pairs of indices `i, j` where
        `ref[i] < ref[j]` but `est[i] >= est[j]`.

    normalizer : float
        The total number of pairs (i, j) under consideration.
        If transitive=True, then this is |{(i,j) : ref[i] < ref[j]}|
        If transitive=False, then this is |{i,j) : ref[i] +1 = ref[j]}|
    '''

    idx = np.argsort(ref)
    ref_sorted = ref[idx]
    est_sorted = est[idx]

    # Find the break-points in ref_sorted
    levels, positions, counts = np.unique(ref_sorted,
                                          return_index=True,
                                          return_counts=True)

    positions = list(positions)
    positions.append(len(ref_sorted))

    index = collections.defaultdict(lambda: slice(0))
    ref_map = collections.defaultdict(lambda: 0)

    for level, cnt, start, end in zip(levels, counts,
                                      positions[:-1], positions[1:]):
        index[level] = slice(start, end)
        ref_map[level] = cnt

    # Now that we have values sorted, apply the inversion-counter to
    # pairs of reference values
    if transitive:
        level_pairs = itertools.combinations(levels, 2)
    else:
        level_pairs = [(i, i+1) for i in levels]

    level_pairs, lcounter = itertools.tee(level_pairs)

    normalizer = float(sum([ref_map[i] * ref_map[j] for (i, j) in lcounter]))

    if normalizer == 0:
        return 0, 0.0

    inversions = 0

    for level_1, level_2 in level_pairs:
        inversions += _count_inversions(est_sorted[index[level_1]],
                                        est_sorted[index[level_2]])

    return inversions, float(normalizer)


def validate_hier_intervals(intervals_hier):
    '''Validate a hierarchical segment annotation.

    Parameters
    ----------
    intervals_hier : ordered list of segmentations

    Raises
    ------
    ValueError
        If any segmentation does not span the full duration of the top-level
        segmentation.

        If any segmentation does not start at 0.
    '''

    # Synthesize a label array for the top layer.
    label_top = util.generate_labels(intervals_hier[0])

    boundaries = set(util.intervals_to_boundaries(intervals_hier[0]))

    for level, intervals in enumerate(intervals_hier[1:], 1):
        # Make sure this level is consistent with the root
        label_current = util.generate_labels(intervals)
        validate_structure(intervals_hier[0], label_top,
                           intervals, label_current)

        # Make sure all previous boundaries are accounted for
        new_bounds = set(util.intervals_to_boundaries(intervals))

        if boundaries - new_bounds:
            warnings.warn('Segment hierarchy is inconsistent '
                          'at level {:d}'.format(level))
        boundaries |= new_bounds


def tmeasure(reference_intervals_hier, estimated_intervals_hier,
             transitive=False, window=15.0, frame_size=0.1, beta=1.0):
    '''Computes the tree measures for hierarchical segment annotations.

    Parameters
    ----------
    reference_intervals_hier : list of ndarray
        ``reference_intervals_hier[i]`` contains the segment intervals
        (in seconds) for the ``i`` th layer of the annotations.  Layers are
        ordered from top to bottom, so that the last list of intervals should
        be the most specific.

    estimated_intervals_hier : list of ndarray
        Like ``reference_intervals_hier`` but for the estimated annotation

    transitive : bool
        whether to compute the t-measures using transitivity or not.

    window : float > 0
        size of the window (in seconds).  For each query frame q,
        result frames are only counted within q +- window.

    frame_size : float > 0
        length (in seconds) of frames.  The frame size cannot be longer than
        the window.

    beta : float > 0
        beta parameter for the F-measure.

    Returns
    -------
    t_precision : number [0, 1]
        T-measure Precision

    t_recall : number [0, 1]
        T-measure Recall

    t_measure : number [0, 1]
        F-beta measure for ``(t_precision, t_recall)``

    Raises
    ------
    ValueError
        If either of the input hierarchies are inconsistent

        If the input hierarchies have different time durations

        If ``frame_size > window`` or ``frame_size <= 0``
    '''

    # Compute the number of frames in the window
    if frame_size <= 0:
        raise ValueError('frame_size ({:.2f}) must be a positive '
                         'number.'.format(frame_size))

    if window is None:
        window_frames = None
    else:
        if frame_size > window:
            raise ValueError('frame_size ({:.2f}) cannot exceed '
                             'window ({:.2f})'.format(frame_size, window))

        window_frames = int(_round(window, frame_size) / frame_size)

    # Validate the hierarchical segmentations
    validate_hier_intervals(reference_intervals_hier)
    validate_hier_intervals(estimated_intervals_hier)

    # Build the least common ancestor matrices
    ref_lca = _lca(reference_intervals_hier, frame_size)
    est_lca = _lca(estimated_intervals_hier, frame_size)

    # Compute precision and recall
    t_recall = _gauc(ref_lca, est_lca, transitive, window_frames)
    t_precision = _gauc(est_lca, ref_lca, transitive, window_frames)

    t_measure = util.f_measure(t_precision, t_recall, beta=beta)

    return t_precision, t_recall, t_measure


def lmeasure(reference_intervals_hier, reference_labels_hier,
             estimated_intervals_hier, estimated_labels_hier,
             frame_size=0.1, beta=1.0):
    '''Computes the tree measures for hierarchical segment annotations.

    Parameters
    ----------
    reference_intervals_hier : list of ndarray
        ``reference_intervals_hier[i]`` contains the segment intervals
        (in seconds) for the ``i`` th layer of the annotations.  Layers are
        ordered from top to bottom, so that the last list of intervals should
        be the most specific.

    reference_labels_hier : list of list of str
        ``reference_labels_hier[i]`` contains the segment labels for the
        ``i``th layer of the annotations

    estimated_intervals_hier : list of ndarray
    estimated_labels_hier : list of ndarray
        Like ``reference_intervals_hier`` and ``reference_labels_hier``
        but for the estimated annotation

    frame_size : float > 0
        length (in seconds) of frames.  The frame size cannot be longer than
        the window.

    beta : float > 0
        beta parameter for the F-measure.

    Returns
    -------
    l_precision : number [0, 1]
        L-measure Precision

    l_recall : number [0, 1]
        L-measure Recall

    l_measure : number [0, 1]
        F-beta measure for ``(l_precision, l_recall)``

    Raises
    ------
    ValueError
        If either of the input hierarchies are inconsistent

        If the input hierarchies have different time durations

        If ``frame_size > window`` or ``frame_size <= 0``
    '''

    # Compute the number of frames in the window
    if frame_size <= 0:
        raise ValueError('frame_size ({:.2f}) must be a positive '
                         'number.'.format(frame_size))

    # Validate the hierarchical segmentations
    validate_hier_intervals(reference_intervals_hier)
    validate_hier_intervals(estimated_intervals_hier)

    # Build the least common ancestor matrices
    ref_meet = _meet(reference_intervals_hier, reference_labels_hier,
                     frame_size)
    est_meet = _meet(estimated_intervals_hier, estimated_labels_hier,
                     frame_size)

    # Compute precision and recall
    l_recall = _gauc(ref_meet, est_meet, True, None)
    l_precision = _gauc(est_meet, ref_meet, True, None)

    l_measure = util.f_measure(l_precision, l_recall, beta=beta)

    return l_precision, l_recall, l_measure


def evaluate(ref_intervals_hier, ref_labels_hier,
             est_intervals_hier, est_labels_hier, **kwargs):
    '''Compute all hierarchical structure metrics for the given reference and
    estimated annotations.

    Examples
    --------
    A toy example with two two-layer annotations

    >>> ref_i = [[[0, 30], [30, 60]], [[0, 15], [15, 30], [30, 45], [45, 60]]]
    >>> est_i = [[[0, 45], [45, 60]], [[0, 15], [15, 30], [30, 45], [45, 60]]]
    >>> ref_l = [ ['A', 'B'], ['a', 'b', 'a', 'c'] ]
    >>> est_l = [ ['A', 'B'], ['a', 'a', 'b', 'b'] ]
    >>> scores = mir_eval.hierarchy.evaluate(ref_i, ref_l, est_i, est_l)
    >>> dict(scores)
    {'T-Measure full': 0.94822745804853459,
     'T-Measure reduced': 0.8732458222764804,
     'T-Precision full': 0.96569179094693058,
     'T-Precision reduced': 0.89939075137018787,
     'T-Recall full': 0.93138358189386117,
     'T-Recall reduced': 0.84857799953694923}

    A more realistic example, using SALAMI pre-parsed annotations

    >>> def load_salami(filename):
    ...     "load SALAMI event format as labeled intervals"
    ...     events, labels = mir_eval.io.load_labeled_events(filename)
    ...     intervals = mir_eval.util.boundaries_to_intervals(events)[0]
    ...     return intervals, labels[:len(intervals)]
    >>> ref_files = ['data/10/parsed/textfile1_uppercase.txt',
    ...              'data/10/parsed/textfile1_lowercase.txt']
    >>> est_files = ['data/10/parsed/textfile2_uppercase.txt',
    ...              'data/10/parsed/textfile2_lowercase.txt']
    >>> ref = [load_salami(fname) for fname in ref_files]
    >>> ref_int = [seg[0] for seg in ref]
    >>> ref_lab = [seg[1] for seg in ref]
    >>> est = [load_salami(fname) for fname in est_files]
    >>> est_int = [seg[0] for seg in est]
    >>> est_lab = [seg[1] for seg in est]
    >>> scores = mir_eval.hierarchy.evaluate(ref_int, ref_lab,
    ...                                      est_hier, est_lab)
    >>> dict(scores)
    {'T-Measure full': 0.66029225561405358,
     'T-Measure reduced': 0.62001868041578034,
     'T-Precision full': 0.66844764668949885,
     'T-Precision reduced': 0.63252297209957919,
     'T-Recall full': 0.6523334654992341,
     'T-Recall reduced': 0.60799919710921635}


    Parameters
    ----------
    ref_intervals_hier : list of list-like
    ref_labels_hier : list of list of str
    est_intervals_hier : list of list-like
    est_labels_hier : list of list of str
        Hierarchical annotations are encoded as an ordered list
        of segmentations.  Each segmentation itself is a list (or list-like)
        of intervals (\*_intervals_hier) and a list of lists of labels
        (\*_labels_hier).

    kwargs
        additional keyword arguments to the evaluation metrics.

    Returns
    -------
    scores :  OrderedDict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.

        T-measures are computed in both the "full" (``transitive=True``) and
        "reduced" (``transitive=False``) modes.

    Raises
    ------
    ValueError
        Thrown when the provided annotations are not valid.
    '''

    # First, find the maximum length of the reference
    _, t_end = _hierarchy_bounds(ref_intervals_hier)

    # Pre-process the intervals to match the range of the reference,
    # and start at 0
    ref_intervals_hier, ref_labels_hier = _align_intervals(ref_intervals_hier,
                                                           ref_labels_hier,
                                                           t_min=0.0,
                                                           t_max=None)

    est_intervals_hier, est_labels_hier = _align_intervals(est_intervals_hier,
                                                           est_labels_hier,
                                                           t_min=0.0,
                                                           t_max=t_end)

    scores = collections.OrderedDict()

    # Force the transitivity setting
    kwargs['transitive'] = False
    (scores['T-Precision reduced'],
     scores['T-Recall reduced'],
     scores['T-Measure reduced']) = util.filter_kwargs(tmeasure,
                                                       ref_intervals_hier,
                                                       est_intervals_hier,
                                                       **kwargs)

    kwargs['transitive'] = True
    (scores['T-Precision full'],
     scores['T-Recall full'],
     scores['T-Measure full']) = util.filter_kwargs(tmeasure,
                                                    ref_intervals_hier,
                                                    est_intervals_hier,
                                                    **kwargs)

    (scores['L-Precision'],
     scores['L-Recall'],
     scores['L-Measure']) = util.filter_kwargs(lmeasure,
                                               ref_intervals_hier,
                                               ref_labels_hier,
                                               est_intervals_hier,
                                               est_labels_hier,
                                               **kwargs)
    return scores



'''
Key Detection involves determining the underlying key (distribution of notes
and note transitions) in a piece of music.  Key detection algorithms are
evaluated by comparing their estimated key to a ground-truth reference key and
reporting a score according to the relationship of the keys.

Conventions
-----------
Keys are represented as strings of the form ``'(key) (mode)'``, e.g. ``'C#
major'`` or ``'Fb minor'``.  The case of the key is ignored.  Note that certain
key strings are equivalent, e.g. ``'C# major'`` and ``'Db major'``.  The mode
may only be specified as either ``'major'`` or ``'minor'``, no other mode
strings will be accepted.

Metrics
-------
* :func:`mir_eval.key.weighted_score`: Heuristic scoring of the relation of two
  keys.
'''

import collections
from . import util

KEY_TO_SEMITONE = {'c': 0, 'c#': 1, 'db': 1, 'd': 2, 'd#': 3, 'eb': 3, 'e': 4,
                   'f': 5, 'f#': 6, 'gb': 6, 'g': 7, 'g#': 8, 'ab': 8, 'a': 9,
                   'a#': 10, 'bb': 10, 'b': 11, 'x': None}


def validate_key(key):
    """Checks that a key is well-formatted, e.g. in the form ``'C# major'``.
   The Key can be 'X' if it is not possible to categorize the Key and mode
   can be 'other' if it can't be categorized as major or minor.

    Parameters
    ----------
    key : str
        Key to verify
    """
    if len(key.split()) != 2 \
            and not (len(key.split()) and key.lower() == 'x'):
        raise ValueError("'{}' is not in the form '(key) (mode)' "
                         "or 'X'".format(key))
    if key.lower() != 'x':
        key, mode = key.split()

        if key.lower() == 'x':
            raise ValueError(
                "Mode {} is invalid; 'X' (Uncategorized) "
                "doesn't have mode".format(mode))
        if key.lower() not in KEY_TO_SEMITONE:
            raise ValueError(
                "Key {} is invalid; should be e.g. D or C# or Eb or "
                "X (Uncategorized)".format(key))
        if mode not in ['major', 'minor', 'other']:
            raise ValueError(
                "Mode '{}' is invalid; must be 'major', 'minor' or 'other'"
                .format(mode))


def validate(reference_key, estimated_key):
    """Checks that the input annotations to a metric are valid key strings and
    throws helpful errors if not.

    Parameters
    ----------
    reference_key : str
        Reference key string.
    estimated_key : str
        Estimated key string.
    """
    for key in [reference_key, estimated_key]:
        validate_key(key)


def split_key_string(key):
    """Splits a key string (of the form, e.g. ``'C# major'``), into a tuple of
    ``(key, mode)`` where ``key`` is is an integer representing the semitone
    distance from C.

    Parameters
    ----------
    key : str
        String representing a key.

    Returns
    -------
    key : int
        Number of semitones above C.
    mode : str
        String representing the mode.
    """
    if key.lower() != 'x':
        key, mode = key.split()
    else:
        mode = None
    return KEY_TO_SEMITONE[key.lower()], mode


def weighted_score(reference_key, estimated_key):
    """Computes a heuristic score which is weighted according to the
    relationship of the reference and estimated key, as follows:

    +------------------------------------------------------+-------+
    | Relationship                                         | Score |
    +------------------------------------------------------+-------+
    | Same key and mode                                    | 1.0   |
    +------------------------------------------------------+-------+
    | Estimated key is a perfect fifth above reference key | 0.5   |
    +------------------------------------------------------+-------+
    | Relative major/minor (same key signature)            | 0.3   |
    +------------------------------------------------------+-------+
    | Parallel major/minor (same key)                      | 0.2   |
    +------------------------------------------------------+-------+
    | Other                                                | 0.0   |
    +------------------------------------------------------+-------+

    Examples
    --------
    >>> ref_key = mir_eval.io.load_key('ref.txt')
    >>> est_key = mir_eval.io.load_key('est.txt')
    >>> score = mir_eval.key.weighted_score(ref_key, est_key)

    Parameters
    ----------
    reference_key : str
        Reference key string.
    estimated_key : str
        Estimated key string.

    Returns
    -------
    score : float
        Score representing how closely related the keys are.
    """
    validate(reference_key, estimated_key)
    reference_key, reference_mode = split_key_string(reference_key)
    estimated_key, estimated_mode = split_key_string(estimated_key)
    # If keys are the same, return 1.
    if reference_key == estimated_key and reference_mode == estimated_mode:
        return 1.
    # If reference or estimated key are x and they are not the same key
    # then the result is 'Other'.
    if reference_key is None or estimated_key is None:
        return 0.
    # If keys are the same mode and a perfect fifth (differ by 7 semitones)
    if (estimated_mode == reference_mode and
            (estimated_key - reference_key) % 12 == 7):
        return 0.5
    # Estimated key is relative minor of reference key (9 semitones)
    if (estimated_mode != reference_mode == 'major' and
            (estimated_key - reference_key) % 12 == 9):
        return 0.3
    # Estimated key is relative major of reference key (3 semitones)
    if (estimated_mode != reference_mode == 'minor' and
            (estimated_key - reference_key) % 12 == 3):
        return 0.3
    # If keys are in different modes and parallel (same key name)
    if estimated_mode != reference_mode and reference_key == estimated_key:
        return 0.2
    # Otherwise return 0
    return 0.


def evaluate(reference_key, estimated_key, **kwargs):
    """Compute all metrics for the given reference and estimated annotations.

    Examples
    --------
    >>> ref_key = mir_eval.io.load_key('reference.txt')
    >>> est_key = mir_eval.io.load_key('estimated.txt')
    >>> scores = mir_eval.key.evaluate(ref_key, est_key)

    Parameters
    ----------
    ref_key : str
        Reference key string.

    ref_key : str
        Estimated key string.

    kwargs
        Additional keyword arguments which will be passed to the
        appropriate metric or preprocessing functions.

    Returns
    -------
    scores : dict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.
    """
    # Compute all metrics
    scores = collections.OrderedDict()

    scores['Weighted Score'] = util.filter_kwargs(
        weighted_score, reference_key, estimated_key)

    return scores



"""
Unit tests for mir_eval.alignment
"""

import glob
import json

import nose.tools
import numpy as np

import mir_eval

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/alignment/ref*.txt"
EST_GLOB = "data/alignment/est*.txt"
SCORES_GLOB = "data/alignment/output*.json"


def __unit_test_alignment_function(metric):
    # Now test validation function
    # alignments must be 1d ndarray
    alignments = np.array([[1.0, 2.0]])
    nose.tools.assert_raises(ValueError, metric, alignments, alignments)
    # alignments must be in seconds, and therefore not negative
    alignments = np.array([-1.0, 2.0])
    nose.tools.assert_raises(ValueError, metric, alignments, alignments)
    # alignments must be sorted
    alignments = np.array([2.0, 1.0])
    nose.tools.assert_raises(ValueError, metric, alignments, alignments)
    # predicted and estimated alignments must have same length
    pred_alignments = np.array([[1.0, 2.0]])
    est_alignments = np.array([[1.0]])
    nose.tools.assert_raises(
        ValueError, metric, est_alignments, pred_alignments
    )


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_alignment_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit tests
    for metric in [
        mir_eval.alignment.absolute_error,
        mir_eval.alignment.percentage_correct,
        mir_eval.alignment.percentage_correct_segments,
        (
            lambda ref_ts, est_ts: mir_eval.alignment.percentage_correct_segments(
                ref_ts, est_ts, duration=max(np.max(ref_ts), np.max(est_ts))
            )
        ),
        mir_eval.alignment.karaoke_perceptual_metric,
    ]:
        yield (__unit_test_alignment_function, metric)
    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, "r") as f:
            expected_scores = json.load(f)
        # Load in an example alignment annotation
        reference_alignments = mir_eval.io.load_events(ref_f)
        # Load in an example alignment tracker output
        estimated_alignments = mir_eval.io.load_events(est_f)
        # Compute scores
        scores = mir_eval.alignment.evaluate(
            reference_alignments, estimated_alignments
        )
        # Compare them
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (
                __check_score,
                sco_f,
                metric,
                scores[metric],
                expected_scores[metric],
            )



#!/usr/bin/env python
'''
Unit tests for mir_eval.tempo
'''
import warnings

import numpy as np
import mir_eval
from nose.tools import raises
import json
import glob


A_TOL = 1e-12


def _load_tempi(filename):

    values = mir_eval.io.load_delimited(filename, [float] * 3)

    return np.concatenate(values[:2]), values[-1][0]


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_zero_tolerance_pass():
    good_ref = np.array([60, 120])
    good_weight = 0.5
    good_est = np.array([120, 180])
    zero_tol = 0.0

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # Try to produce the warning
        mir_eval.tempo.detection(good_ref, good_weight, good_est, tol=zero_tol)

        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == 'A tolerance of 0.0 may ' \
                                     'not lead to the results you expect.'


def test_tempo_pass():
    good_ref = np.array([60, 120])
    good_weight = 0.5
    good_est = np.array([120, 180])
    good_tol = 0.08

    for good_tempo in [np.array([50, 50]), np.array([0, 50]),
                       np.array([50, 0])]:
        yield mir_eval.tempo.detection, good_tempo,\
            good_weight, good_est, good_tol
        yield mir_eval.tempo.detection, good_ref,\
            good_weight, good_tempo, good_tol

    # allow both estimates to be zero
    yield mir_eval.tempo.detection, good_ref,\
        good_weight, np.array([0, 0]), good_tol


def test_tempo_fail():

    @raises(ValueError)
    def __test(ref, weight, est, tol):
        mir_eval.tempo.detection(ref, weight, est, tol=tol)

    good_ref = np.array([60, 120])
    good_weight = 0.5
    good_est = np.array([120, 180])
    good_tol = 0.08

    for bad_tempo in [np.array([-1, -1]), np.array([-1, 0]),
                      np.array([-1, 50]), np.array([0, 1, 2]), np.array([0])]:
        yield __test, bad_tempo, good_weight, good_est, good_tol
        yield __test, good_ref, good_weight, bad_tempo, good_tol

    for bad_weight in [-1, 1.5]:
        yield __test, good_ref, bad_weight, good_est, good_tol

    for bad_tol in [-1, 1.5]:
        yield __test, good_ref, good_weight, good_est, bad_tol

    # don't allow both references to be zero
    yield __test, np.array([0, 0]), good_weight, good_ref, good_tol


def test_tempo_regression():
    REF_GLOB = 'data/tempo/ref*.lab'
    EST_GLOB = 'data/tempo/est*.lab'
    SCORES_GLOB = 'data/tempo/output*.json'

    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files)

    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as fdesc:
            expected_scores = json.load(fdesc)

        ref_tempi, ref_weight = _load_tempi(ref_f)
        est_tempi, _ = _load_tempi(est_f)

        scores = mir_eval.tempo.evaluate(ref_tempi, ref_weight, est_tempi)

        for metric in scores:
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])



# CREATED: 4/15/14 9:42 AM by Justin Salamon <justin.salamon@nyu.edu>
'''
Unit tests for mir_eval.melody
'''

import numpy as np
import json
import nose.tools
import mir_eval
import glob
import warnings

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/melody/ref*.txt'
EST_GLOB = 'data/melody/est*.txt'
SCORES_GLOB = 'data/melody/output*.json'


def test_hz2cents():
    # Unit test some simple values
    hz = np.array([0., 10., 5., 320., 1420.31238974231])
    # Expected cent conversion
    expected_cent = np.array([0., 0., -1200., 6000., 8580.0773605])
    assert np.allclose(mir_eval.melody.hz2cents(hz), expected_cent)


def test_freq_to_voicing():
    # Unit test some simple values
    hz = np.array([0., 100., -132.])
    expected_hz = np.array([0., 100., 132.])
    expected_voicing = np.array([0, 1, 0])
    # Check voicing conversion
    res_hz, res_voicing = mir_eval.melody.freq_to_voicing(hz)
    assert np.all(res_hz == expected_hz)
    assert np.all(res_voicing == expected_voicing)

    # Unit test some simple values where voicing is given
    hz = np.array([0., 100., -132., 0, 131.])
    voicing = np.array([0.8, 0.0, 1.0, 0.0, 0.5])
    expected_hz = np.array([0., 100., 132., 0., 131.])
    expected_voicing = np.array([0.0, 0.0, 1.0, 0.0, 0.5])
    # Check voicing conversion
    res_hz, res_voicing = mir_eval.melody.freq_to_voicing(hz, voicing=voicing)
    assert np.all(res_hz == expected_hz)
    assert np.all(res_voicing == expected_voicing)


def test_constant_hop_timebase():
    hop = .1
    end_time = .35
    expected_times = np.array([0, .1, .2, .3])
    res_times = mir_eval.melody.constant_hop_timebase(hop, end_time)
    assert np.allclose(res_times, expected_times)


def test_resample_melody_series():
    # Check for a small example including a zero transition
    times = np.arange(4)/35.0
    cents = np.array([2., 0., -1., 1.])
    voicing = np.array([1, 0, 1, 1])
    times_new = np.linspace(0, .08, 9)
    expected_cents = np.array([2., 2., 2., 0., 0., 0., -.8, -.1, .6])
    expected_voicing = np.array([1, 1, 1, 0, 0, 0, 1, 1, 1])
    (res_cents,
     res_voicing) = mir_eval.melody.resample_melody_series(times, cents,
                                                           voicing, times_new)
    assert np.allclose(res_cents, expected_cents)
    assert np.allclose(res_voicing, expected_voicing)

    # Check for a small example including a zero transition - nonbinary voicing
    times = np.arange(4)/35.0
    cents = np.array([2., 0., -1., 1.])
    voicing = np.array([0.8, 0.0, 0.2, 1.0])
    times_new = np.linspace(0, .08, 9)
    expected_cents = np.array([2., 2., 2., 0., 0., 0., -.8, -.1, .6])
    expected_voicing = np.array(
        [0.8, 0.52, 0.24, 0.01, 0.08, 0.15, 0.28, 0.56, 0.84]
    )
    (res_cents,
     res_voicing) = mir_eval.melody.resample_melody_series(times, cents,
                                                           voicing, times_new)
    assert np.allclose(res_cents, expected_cents)
    assert np.allclose(res_voicing, expected_voicing)


def test_resample_melody_series_same_times():
    # Check the case where the time bases are identical
    times = np.array([0.0, 0.1, 0.2, 0.3])
    times_new = np.array([0.0, 0.1, 0.2, 0.3])
    cents = np.array([2., 0., -1., 1.])
    voicing = np.array([0, 0, 1, 1])
    expected_cents = np.array([2., 0., -1., 1.])
    expected_voicing = np.array([False, False, True, True])
    (res_cents,
     res_voicing) = mir_eval.melody.resample_melody_series(times, cents,
                                                           voicing, times_new)
    assert np.allclose(res_cents, expected_cents)
    assert np.allclose(res_voicing, expected_voicing)

    # Check the case where the time bases are identical - nonbinary voicing
    times = np.array([0.0, 0.1, 0.2, 0.3])
    times_new = np.array([0.0, 0.1, 0.2, 0.3])
    cents = np.array([2., 0., -1., 1.])
    voicing = np.array([0.5, 0.8, 0.9, 1.0])
    expected_cents = np.array([2., 0., -1., 1.])
    expected_voicing = np.array([0.5, 0.8, 0.9, 1.0])
    (res_cents,
     res_voicing) = mir_eval.melody.resample_melody_series(times, cents,
                                                           voicing, times_new)
    assert np.allclose(res_cents, expected_cents)
    assert np.allclose(res_voicing, expected_voicing)


def test_to_cent_voicing():
    # We'll just test a few values from one of the test annotations
    ref_file = sorted(glob.glob(REF_GLOB))[0]
    ref_time, ref_freq = mir_eval.io.load_time_series(ref_file)
    est_file = sorted(glob.glob(EST_GLOB))[0]
    est_time, est_freq = mir_eval.io.load_time_series(est_file)
    ref_v, ref_c, est_v, est_c = mir_eval.melody.to_cent_voicing(ref_time,
                                                                 ref_freq,
                                                                 est_time,
                                                                 est_freq)
    # Expected values
    test_range = np.arange(220, 225)
    expected_ref_v = np.array([False, False, False, True, True])
    expected_ref_c = np.array([0., 0., 0., 6056.8837818916609,
                               6028.5504583021921])
    expected_est_v = np.array([False]*5)
    expected_est_c = np.array([5351.3179423647571]*5)
    assert np.allclose(ref_v[test_range], expected_ref_v)
    assert np.allclose(ref_c[test_range], expected_ref_c)
    assert np.allclose(est_v[test_range], expected_est_v)
    assert np.allclose(est_c[test_range], expected_est_c)

    # Test that a 0 is added to the beginning
    for return_item in mir_eval.melody.to_cent_voicing(
            np.array([1., 2.]), np.array([440., 442.]), np.array([1., 2.]),
            np.array([441., 443.])):
        assert len(return_item) == 3
        assert return_item[0] == return_item[1]

    # Test custom voicings
    ref_time, ref_freq = mir_eval.io.load_time_series(ref_file)
    _, ref_reward = mir_eval.io.load_time_series("data/melody/reward00.txt")
    _, est_voicing = mir_eval.io.load_time_series(
        "data/melody/voicingest00.txt"
    )
    (ref_v, ref_c,
     est_v, est_c) = mir_eval.melody.to_cent_voicing(ref_time,
                                                     ref_freq,
                                                     est_time,
                                                     est_freq,
                                                     est_voicing=est_voicing,
                                                     ref_reward=ref_reward)
    # Expected values
    test_range = np.arange(220, 225)
    expected_ref_v = np.array([0., 0., 0., 1., 0.3])
    expected_ref_c = np.array([0., 0., 0., 6056.8837818916609,
                               6028.5504583021921])
    expected_est_v = np.array([0.2, 0.2, 0.2, 0.2, 0.2])
    expected_est_c = np.array([5351.3179423647571]*5)
    assert np.allclose(ref_v[test_range], expected_ref_v)
    assert np.allclose(ref_c[test_range], expected_ref_c)
    assert np.allclose(est_v[test_range], expected_est_v)
    assert np.allclose(est_c[test_range], expected_est_c)


def test_continuous_voicing_metrics():
    ref_time = np.array([0.0, 0.1, 0.2, 0.3])
    ref_freq = np.array([440.0, 0.0, 220.0, 220.0])

    est_time = np.array([0.0, 0.1, 0.2, 0.3])
    est_freq = np.array([440.1, 330.0, 440.0, 330.0])

    # test different estimate voicings
    all_est_voicing = [
        np.array([1.0, 0.0, 1.0, 1.0]),  # perfect
        np.array([0.0, 1.0, 0.0, 0.0]),  # all wrong
        np.array([0.5, 0.5, 0.5, 0.5]),  # all 0.5
        np.array([0.8, 0.2, 0.8, 0.8]),  # almost right
        np.array([0.2, 0.8, 0.2, 0.2]),  # almost wrong
    ]

    all_expected = [
        # perfect
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.0,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.5,
        },
        # all wrong
        {
            'Voicing Recall': 0.0,
            'Voicing False Alarm': 1.0,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.0,
        },
        # all 0.5
        {
            'Voicing Recall': 0.5,
            'Voicing False Alarm': 0.5,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.25,
        },
        # almost right
        {
            'Voicing Recall': 0.8,
            'Voicing False Alarm': 0.2,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.4,
        },
        # almost wrong
        {
            'Voicing Recall': 0.2,
            'Voicing False Alarm': 0.8,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.1,
        },
    ]

    for est_voicing, expected_scores in zip(all_est_voicing, all_expected):
        actual_scores = mir_eval.melody.evaluate(ref_time, ref_freq, est_time,
                                                 est_freq,
                                                 est_voicing=est_voicing)
        for metric in actual_scores:
            assert np.isclose(actual_scores[metric], expected_scores[metric])

    # test different rewards
    all_rewards = [
        np.array([0.5, 0.5, 0.5, 0.5]),  # uniform
        np.array([0.3, 0.3, 0.3, 0.3]),  # uniform - different number
        np.array([0.0, 0.0, 0.0, 0.0]),  # all zero
        np.array([1.0, 0.0, 0.0, 0.0]),  # one weight
        np.array([1.0, 0.0, 1.0, 0.0]),  # two weights
        np.array([1.0, 0.0, 0.5, 0.5]),  # slightly generous
        np.array([0.1, 0.0, 0.1, 0.8]),  # big penalty
    ]
    est_voicing = np.array([1.0, 0.0, 1.0, 1.0])

    all_expected = [
        # uniform
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.0,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.5,
        },
        # uniform - different number
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.0,
            'Raw Pitch Accuracy': 1. / 3.,
            'Raw Chroma Accuracy': 2. / 3.,
            'Overall Accuracy': 0.5,
        },
        # all zero
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.75,
            'Raw Pitch Accuracy': 0.0,
            'Raw Chroma Accuracy': 0.0,
            'Overall Accuracy': 0.25,
        },
        # one weight
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 2. / 3.,
            'Raw Pitch Accuracy': 1.0,
            'Raw Chroma Accuracy': 1.0,
            'Overall Accuracy': 0.5,
        },
        # two weights
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.5,
            'Raw Pitch Accuracy': 0.5,
            'Raw Chroma Accuracy': 1.0,
            'Overall Accuracy': 0.5,
        },
        # slightly generous
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.0,
            'Raw Pitch Accuracy': 0.5,
            'Raw Chroma Accuracy': 0.75,
            'Overall Accuracy': 0.625,
        },
        # big penalty
        {
            'Voicing Recall': 1.0,
            'Voicing False Alarm': 0.0,
            'Raw Pitch Accuracy': 0.1,
            'Raw Chroma Accuracy': 0.2,
            'Overall Accuracy': 0.325,
        },
    ]

    for ref_reward, expected_scores in zip(all_rewards, all_expected):
        actual_scores = mir_eval.melody.evaluate(ref_time, ref_freq, est_time,
                                                 est_freq,
                                                 est_voicing=est_voicing,
                                                 ref_reward=ref_reward)
        for metric in actual_scores:
            assert np.isclose(actual_scores[metric], expected_scores[metric])


def __unit_test_voicing_measures(metric):
    # We need a special test for voicing_measures because it only takes 2 args
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for warnings due to empty voicing arrays
        score = metric(np.array([]), np.array([]))
        assert len(w) == 4
        assert np.all([issubclass(wrn.category, UserWarning) for wrn in w])
        assert [str(wrn.message)
                for wrn in w] == ["Reference voicing array is empty.",
                                  "Estimated voicing array is empty.",
                                  "Reference melody has no voiced frames.",
                                  "Estimated melody has no voiced frames."]
        # And that the metric is 0
        assert np.allclose(score, 0)
        # Also test for a warning when the arrays have non-voiced content
        metric(np.ones(10), np.zeros(10))
        assert len(w) == 5
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Estimated melody has no voiced frames."

    # Now test validation function - voicing arrays must be the same size
    nose.tools.assert_raises(ValueError, metric, np.ones(10), np.ones(12))


def __unit_test_melody_function(metric):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for warnings due to empty voicing arrays
        score = metric(np.array([]), np.array([]), np.array([]), np.array([]))
        assert len(w) == 6
        assert np.all([issubclass(wrn.category, UserWarning) for wrn in w])
        assert [str(wrn.message)
                for wrn in w] == ["Reference voicing array is empty.",
                                  "Estimated voicing array is empty.",
                                  "Reference melody has no voiced frames.",
                                  "Estimated melody has no voiced frames.",
                                  "Reference frequency array is empty.",
                                  "Estimated frequency array is empty."]
        # And that the metric is 0
        assert np.allclose(score, 0)
        # Also test for a warning when the arrays have non-voiced content
        metric(np.ones(10), np.arange(10), np.zeros(10), np.arange(10))
        assert len(w) == 7
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Estimated melody has no voiced frames."

    # Now test validation function - all inputs must be same length
    nose.tools.assert_raises(ValueError, metric, np.ones(10),
                             np.ones(12), np.ones(10), np.ones(10))


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_melody_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit tests
    for metric in [mir_eval.melody.voicing_measures,
                   mir_eval.melody.raw_pitch_accuracy,
                   mir_eval.melody.raw_chroma_accuracy,
                   mir_eval.melody.overall_accuracy]:
        if metric == mir_eval.melody.voicing_measures:
            yield (__unit_test_voicing_measures, metric)
        else:
            yield (__unit_test_melody_function, metric)
    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in reference melody
        ref_time, ref_freq = mir_eval.io.load_time_series(ref_f)
        # Load in estimated melody
        est_time, est_freq = mir_eval.io.load_time_series(est_f)
        scores = mir_eval.melody.evaluate(ref_time, ref_freq, est_time,
                                          est_freq)
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])


def test_melody_functions_continuous_voicing_equivalence():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit tests
    for metric in [mir_eval.melody.voicing_measures,
                   mir_eval.melody.raw_pitch_accuracy,
                   mir_eval.melody.raw_chroma_accuracy,
                   mir_eval.melody.overall_accuracy]:
        if metric == mir_eval.melody.voicing_measures:
            yield (__unit_test_voicing_measures, metric)
        else:
            yield (__unit_test_melody_function, metric)
    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in reference melody
        ref_time, ref_freq = mir_eval.io.load_time_series(ref_f)
        ref_reward = np.ones(ref_time.shape)  # uniform reward
        # Load in estimated melody
        est_time, est_freq = mir_eval.io.load_time_series(est_f)
        # voicing equivalent from frequency
        est_voicing = (est_freq >= 0).astype('float')
        scores = mir_eval.melody.evaluate(ref_time, ref_freq, est_time,
                                          est_freq, est_voicing=est_voicing,
                                          ref_reward=ref_reward)
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])



from setuptools import setup

with open('README.rst') as file:
    long_description = file.read()

setup(
    name='mir_eval',
    version='0.7',
    description='Common metrics for common audio/music processing tasks.',
    author='Colin Raffel',
    author_email='craffel@gmail.com',
    url='https://github.com/craffel/mir_eval',
    packages=['mir_eval'],
    long_description=long_description,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        'Development Status :: 5 - Production/Stable',
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
    ],
    keywords='audio music mir dsp',
    license='MIT',
    install_requires=[
        'numpy >= 1.7.0',
        'scipy >= 1.0.0',
        'future',
        'six'
    ],
    extras_require={
        'display': ['matplotlib>=1.5.0',
                    'scipy>=1.0.0'],
        'testing': ['matplotlib>=2.1.0,<3']
    }
)
